//****************************************************************************
// File: backgrnd.c
//
// Purpose: Contains main message loop and application entry point
//
// Functions:
//    WinMain() - initializes everything and enters message loop
//
// Development Team: Paul Stafford (Original Author)
//                   Krishna Nareddy (Modified sample to confirm to new coding
//                                    standards )
//
// BACKGRND uses a PeekMessage() loop to do background processing.
// The background task is to draw random rectangles. One "batch" of
// rectangles is drawn each time through the PeekMessage() loop,
// and all of the rectangles in each batch are drawn the same color.
// This provides a visual indication of how much background work is
// being done before BACKGRND offers to yield.
//  Both the "batch size" and the total number of rectangles to be
// drawn can be changed via the menu.
//
// BACKGRND does not replace the main GetMessage() loop with
// a PeekMessage() loop. Instead, it enters a PeekMessage() loop
// only when the user chooses the "Draw!" menu item.
//
// While drawing the rectangles, BACKGRND puts up a modeless dialog
// box that:
//
// 1. allows the user to cancel the background processing before
//    all of the rectangles have been drawn (by hitting the "Cancel"
//    button)
// 2. indicates both the total number of rectangles to be drawn
//    and the batch size
// 3. displays a "gas gauge" that indicates the percentage
//    of the background job that has been completed
//
// BACKGRND can be run along with another Windows application
// (say NOTEPAD) to illustrate the difference between "polite"
// and "impolite" background processing.
//
//  If the batch size is set to a "polite" small value (say 10
// rectangles each time through the PeekMessage() loop), the user can
// easily type within NOTEPAD while BACKGRND is drawing rectangles.
// BACKGRND draws rectangles in the spare time slices between
// the messages generated by the user's keystrokes, while calling
// PeekMessage() often enough that the user will still find
// NOTEPAD to be responsive.
//
// If the batch size is set to an "impolite" large value (say 1000
// rectangles each time through the PeekMessage() loop), BACKGRND can
// complete its background job much more quickly, because there are
// fewer calls to PeekMessage() and thus less overhead. However,
// a user trying to type into NOTEPAD will find the keyboard so
// unresponsive that NOTEPAD will be essentially useless.
//
//
// MENU CHOICES:
//
//   Rectangles/Total     - set total number of rectangles to be drawn
//   Rectangles/At A Time - set number of rectangles to be drawn each
//                          time through the PeekMessage()
//                          loop - the 'batch size'
//   Draw!                - start background processing
//
//
// Written by Microsoft Product Support Services, Windows Developer Support
// Copyright (c) 1992 Microsoft Corporation. All rights reserved.
//****************************************************************************

#include <windows.h>
#include "global.h"
#include "init.h"
#include "backgrnd.h"

//***********************************************************************
// Function: WinMain
//
// Purpose: Called by Windows on app startup.  Initializes everything,
//          and enters a message loop.
//
// Parameters:
//    hInstance     == Handle to _this_ instance.
//    hPrevInstance == Handle to last instance of app.
//    lpCmdLine     == Command Line passed into app.
//    nCmdShow      == How app should come up (i.e. minimized/normal)
//
// Returns: Return value from PostQuitMessage.
//
// Comments:
//
// History:  Date       Author        Reason
//           3/5/92     Krishna       Created
//****************************************************************************

int PASCAL WinMain (HANDLE hInstance,  // This instance
                    HANDLE hPrevInstance,   // Last instance
                    LPSTR lpCmdLine,   // Command Line
                    int nCmdShow) // Minimized or Normal?
{
   MSG msg;

   if (!hPrevInstance)
      if (!InitApplication(hInstance))
         return (FALSE);

   if (!InitInstance(hInstance, nCmdShow))
      return (FALSE);

   while (GetMessage(&msg,   // Put Message Here
                     NULL,   // Handle of window receiving msg
                     NULL,   // lowest message to examine
                     NULL))  // highest message to examine
   {
      TranslateMessage(&msg);     // Translates virtual key codes
      DispatchMessage(&msg); // Dispatches message to window
   }

   return (msg.wParam); // Returns the value from PostQuitMessage
}
