
/* A bar graph demonstration program. */

#include "dos.h"
#include "stdio.h"
#include "conio.h"
#include "ctype.h"
#include "stdlib.h"
#include "math.h"
#include "string.h"

#define XMAX 319
#define YMAX 199

void bargraph(double *data, int num, int offset,
              int min, int max, int width);
void grid(int min, int max);
void label(char str[][20], int num);
void hashlines(void), legend(char names[][20], int num);

void line(int startx, int starty, int endx, int endy, 
         int color);
void fill_box(int startx, int starty, int endx, int endy, 
         int color);
void mempoint(int x, int y, int color);
void palette(int pnum);
void mode(int mode_code);
void goto_xy(int x, int y);

main(void)
{
  double widget[]={
    10.1, 20, 30, 35.34, 50
  };
  double global[]={
    19, 20, 8.8, 30, 40
  };
  double tower[]={
    25.25, 19, 17.4, 33, 29
  };

  char n[][20]={
    "Widget", 
    "Global",
    "Tower"    
  };
  char lab[][20]={
    "1987",
    "1988",
    "1989",
    "1990",
    "1991"
  };

  mode(4); /* 320x200  graphics mode */
  palette(0);

  grid(0, 50); /* display the grid */
  hashlines(); /* display dotted reference lines */
  label(lab, 5); /* label the graph */
  legend(n, 3); /* show the legend */

  /* plot the stock prices of the three companies */
  bargraph(widget, 5, 0, 0, 50, 4);
  bargraph(global, 5, 10, 0, 50, 4);
  bargraph(tower, 5, 20, 0, 50, 4);

  getch();
  mode(3);

  return 0;
}

/* Draw the chart grid. */
void grid(int min, int max)
{
  goto_xy(0, 22); printf("%d", min);
  goto_xy(0, 0); printf("%d", max);
  line(10, 180, 300, 180, 1);
}

/* Display labels on the screen. */
void label(char str[][20], /* strings to display */
           int num) /* number of labels */
{
  int i, j, inc;

  inc = 38 / num;
  i = 2; /* initial horizontal starting point */
  for(j=0; j<num; j++) {
    goto_xy(i, 23);
    printf(str[j]);
    i += inc;
  }
}

/* Draw dotted reference lines across the screen. */
void hashlines(void)
{
  int i, j;

  for(i=10; i<300; i+=5) {
    for(j=10; j<180; j+=10)
      mempoint(i, j, 3); /* one point every 5 pixels */
  }
}

/* Show a legend. */
void legend(char names[][20],
            int num) /* number of names */
{
  int color=1, i, j;

  goto_xy(0, 24); /* legend goes on bottom line */
  j = 0;
  for(i=0; i<num; i++) {
    /* print the label */
    printf("%s    ", names[i]);

    /* Compute where colored box goes by
       converting cursor location into graphics
       coordinates.  Each character is 8 pixels wide in
       mode 4 graphics.
    */
    j += strlen(names[i])*8+4;
    fill_box(j, 192, j+12, 198, color);
    j += 28;  /* advance to next field */
    color++;
    if(color>3) color = 1;
  }
}    
   
/* Display a bar chart. */
void bargraph(double *data, /* array of data */
     int num, /* number of elements in array */
     int offset, /* determine exact screen position */
     int min, int max, /* min and max values to be plotted */
     int width) /* thickness of the lines */
{
  int y, t, incr;
  double norm_data, norm_ratio, spread;
  static int color=0;
  int tempwidth;
 
  /* always use a different color */
  color++;
  if(color>3) color = 1;

  /* determine normalization factor */
  spread = (double) max-min; 
  norm_ratio = 180/spread;  

  incr = 280/num; /* determine spacing between lines */
  tempwidth = width;
  for(t=0; t<num; ++t) {
    norm_data = data[t];

    /* adjust for negative values */
    norm_data = norm_data-(double) min;  

    norm_data *= norm_ratio; /* normalize */
    y = (int) norm_data; /* type conversion */
    do { 
      line(((t*incr)+20+offset+width), 179,  
        ((t*incr)+20+offset+width), 179-y, color);
      width--;
    } while (width);
    width = tempwidth;
  }
}

/* Draw a line in specified color
   using Bresenham's integer based algorithm.
*/
void line(int startx, int starty, /* upper left corner */
          int endx, int endy, /* lower right corner */
          int color) /* color */
{
  register int t, distance;
  int x=0, y=0, delta_x, delta_y;
  int incx, incy;

  /* Compute the distances in both directions. */
  delta_x = endx-startx;
  delta_y = endy-starty;

  /* Compute the direction of the increment.
     An increment of 0 means either a vertical
     or horizontal line.
  */
  if(delta_x>0) incx = 1;
  else if(delta_x==0) incx = 0;
  else incx = -1;

  if(delta_y>0) incy = 1;
  else if(delta_y==0) incy = 0;
  else incy = -1;

  /* determine which distance is greater */
  delta_x = abs(delta_x);
  delta_y = abs(delta_y);
  if(delta_x>delta_y) distance = delta_x;
  else distance = delta_y;

  /* draw the line */
  for(t=0; t<=distance+1; t++) {
    mempoint(startx, starty, color);
    x += delta_x;
    y += delta_y;  
    if(x>distance) {
      x -= distance;
      startx += incx;
    }
    if(y>distance) {
      y -= distance;
      starty += incy;
    }
  }
}

/* Fill box with specified color. */
void fill_box(int startx, int starty, /* upper left corner */
          int endx, int endy, /* lower right corner */
          int color) /* color */
{
  register int i, begin, end;

  begin = startx<endx ? startx : endx;
  end = startx>endx ? startx : endx;

  for(i=begin; i<=end;i++)
    line(i, starty, i, endy, color);
}

/* Write a point directly to mode 4 graphics RAM. */
void mempoint(int x, int y, /* coordinate of point */
              int color) /* color */
{
  union mask {
    char c[2];
    int i;
  } bit_mask;
  int index, bit_position;
  unsigned char t;
  char xor; /* xor color in or overwrite */

  /* pointer to mode 4 graphics RAM */
  char far *ptr = (char far *) 0xB8000000; 

  bit_mask.i = 0xFF3F;  /* 11111111 00111111 in binary */

  /* check range */
  if(x<0 || x>XMAX || y<0 || y>YMAX) return;

  xor = color & 128; /* see if xor mode is set */
  color = color & 127; /* mask off high bit */
 
  /* set bit_mask and color bits to the right location */
  bit_position = x%4;
  color <<= 2*(3-bit_position);
  bit_mask.i >>= 2*bit_position;


  /* find the correct byte in screen memory */
  index = y*40 + (x >> 2);
  if(y % 2) index += 8152; /* if odd use 2nd bank */

  /* write the color */
  if(!xor) { /* overwrite mode */
    t = *(ptr+index) & bit_mask.c[0];
    *(ptr+index) = t | color;
  }
  else { /* xor mode */
    t = *(ptr+index) | (char) 0;
    *(ptr+index) = t ^ color;
  }
}

/* Set the palette. */
void palette(int pnum)
{
  union REGS r;

  r.h.bh = 1;   /* code for palette select */
  r.h.bl = pnum;
  r.h.ah = 11;  /* set palette function */
  int86(0x10, &r, &r);
}

/* Set the video mode. */
void mode(int mode_code)
{
  union REGS r;

  r.h.al = mode_code;
  r.h.ah = 0;
  int86(0x10, &r, &r);
}

/* Send the cursor to the specified X,Y position. */
void goto_xy(int x, int y)
{
  union REGS r;

  r.h.ah = 2; /* cursor addressing function */
  r.h.dl = x; /* column coordinate */
  r.h.dh = y; /* row coordinate */
  r.h.bh = 0; /* video page */
  int86(0x10, &r, &r);
}

