////////////////////////////////////////////////////////////
// Btree.h: A file-based btree class
// Copyright(c) 1993 Azarona Software. All rights reserved.
////////////////////////////////////////////////////////////
#ifndef H_BTREE
#define H_BTREE

#include "fmgr.h"
#include "mnodeptr.h"

// Function return status codes

const int NODE_OVERFLOW =  2;
const int SUCCESS       =  1;
const int FAIL          =  0;
const int DUPLKEY       = -1;
const int ALLOCERR      = -2;

struct BtreeHeader { // Stored with every tree
  long root_addr;
  unsigned order;
  unsigned num_entries;
  unsigned num_nodes;
  int height;
};

class Btree {  // Btree file class
protected:
 
  FmgrPtr f;          // File the Btree is connected to
  long bh_addr;       // Address of the Btree header
  BtreeHeader bh;     // Btree header
  MnodeCache  cache; // Node cache
  MnodePtr root;  // Pointer to the root node

  void ReadHdr(); 
  void WriteHdr();

  int Insert(Entry &e, MnodePtr &t);
  int Delete(Entry &e, MnodePtr &p);
  void RestoreBalance(MnodePtr &p, int posn);
  void RotateRight(MnodePtr &p, int posn);
  void RotateLeft(MnodePtr &p, int posn);
  void Merge(MnodePtr &p, int posn);

  static void PrintNode(MnodePtr n);
  static void PrintTree(MnodePtr t, int sp);

public:

  Btree(int cs = 8);
  ~Btree();

  // File management routines

  int Connect(FmgrPtr &fptr, int create, long bh_addr=sizeof(FmgrHeader));
  void Disconnect();
  int Create(char *fname, long bh_addr=sizeof(FmgrHeader));
  int Open(char *fname, Fmgr::AccessMode mode, 
    long bh_addr=sizeof(FmgrHeader));
  void Flush(int clear=0);
  void Close();

  int Search(Entry &e);
  int FullSearch(const Entry &e);
  int Add(char *k, long d=0);
  int Remove(char *k, long d = 0);

  int IsEmpty() const;

  int IsOpen() const;
  int OK() const;
  int operator!() const;
  operator int() const;
  void ClearErr();

  void Statistics(int full);
  void PrintTree();

};

inline Btree::~Btree()
{
  Disconnect();
}

inline void Btree::Close()
{
  Disconnect();
}

inline int Btree::IsEmpty() const
{
  return bh.num_entries == 0;
}

inline int Btree::IsOpen() const
{
  return f->IsOpen();
}

inline int Btree::OK() const
{
  return f->OK();
}

inline int Btree::operator!() const
{
  return f->operator!();
}

inline Btree::operator int() const
{
  return f->operator int();
}

inline void Btree::ClearErr()
{
  f->ClearErr();
}

inline void Btree::PrintTree()
{
  printf("\nTree dump:\n");
  PrintTree(root, 0);
}

#endif
