/*
    file.c

    File open/close functions


*/

#include "global.h"

//
// local defines
//

// <dwFlags> for PromptForFileName()
#define PFFN_OPENFILE        0x0001    // prompt to open a file
#define PFFN_SAVEFILE        0x0002    // prompt to save a file
#define PFFN_SHOWDEFAULT     0x0004    // init. show what's in <achFileName>
#define PFFN_OVERWRITEPROMPT 0x0008    // prompt the user about overwriting
#define PFFN_UPPERCASE       0x0010    // make the name uppercase

//
// local data
//

static char *szUntitled = "(Untitled)";
static char *szOpenFilter = "Text Files (*.txt)\0*.txt\0"
                            "All Files (*.*)\0*.*\0"
                            "\0";

//
// local functions
//

BOOL PromptForFileName(HWND hwndOwner,   // window that will own dialog box
                       HANDLE hInst,     // module that contains the resources
                       LPSTR achFileName,// where to put file name
                       WORD cchFileName, // size of <achFileName>
                       LPSTR szCaption,  // caption string
                       LPSTR szFilter,   // file filter
                       LPSTR szDefExt,   // default extension
                       DWORD dwFlags);   // random flags


/* fOK = FileSave(fSaveAs)
 *
 * Do a File/Save operation (if <fSaveAs> is FALSE) or a File/SaveAs
 * operation (if <fSaveAs> is TRUE).  Return TRUE unless the user cancelled
 * or the operation failed.
 */
BOOL FileSave(WORD wFlags)
{
    BOOL        fOK = TRUE;    // function succeeded?
    int        fh = -1;    // DOS file handle
    char        achFileName[_MAX_PATH]; // user-specified file name
    HCURSOR        hcurPrev = NULL; // cursor before hourglass
    long        lFileLength;    // length of the file
    LPSTR        pFileImage;    // memory block to hold file image


    pFileImage = CreateTextImage();
    lFileLength = lstrlen(pFileImage);


    if ((wFlags & FILE_SAVE_AS))
    {
        /* set up the default file name */
        lstrcpy(achFileName, "RIPWATCH.TXT");

        /* prompt the user for the name of the file to save to */
        if (!PromptForFileName(hwndMain, 
                               hAppInstance,
                               achFileName,
                               sizeof(achFileName), 
                               "Save File",
                               szOpenFilter,
                               "txt",
                               PFFN_SAVEFILE | PFFN_SHOWDEFAULT |
                                PFFN_OVERWRITEPROMPT | PFFN_UPPERCASE))
            goto RETURN_ERROR;

    }

    /* show hourglass cursor */
    hcurPrev = SetCursor(LoadCursor(NULL, IDC_WAIT));

    /* open the file -- if it already exists, truncate it to zero bytes */
    if ((fh = _lcreat(achFileName, 0)) < 0)
        goto ERROR_SAVING;

    /* write the file... */

    /* write the file (at most 64K bytes) */
    if ((long) _lwrite(fh, pFileImage, (int) lFileLength) != lFileLength)
        goto ERROR_SAVING;

    goto RETURN_SUCCESS;

ERROR_SAVING:                // display generic error message

    Message(0, "Error saving %s", (LPSTR) achFileName);
    goto RETURN_ERROR;

RETURN_ERROR:                // do error exit without error message

    fOK = FALSE;

RETURN_SUCCESS:                // normal exit

    if (hcurPrev != NULL)
        SetCursor(hcurPrev);

    if (fh >= 0)
        _lclose(fh);
    if (fOK)

    DeleteTextImage(pFileImage);

    return fOK;
}


/* fOK = PromptForFileName(hwndOwner, hInst, achFileName, cchFileName,
 *                         idCaption, idFilter, idDefExt, dwFlags)
 *
 * Prompt the user for the name of a file to open or save to.  <hwndOwner>
 * is the window that will own the dialog.  <hInst> is the module that the
 * string resources and RCDATA resource (see below) will be loaded from.
 *
 * The returned file name will be placed in <achFileName>, which must
 * have a capacity of at least <cchFileName> bytes.
 *
 * <idCaption> is the string resource id of the caption to display.
 * <idFilter> is the RCDATA resource id that contains the filter spec
 * (in the format used by COMMDLG).  <idDefExt> is the string resource id
 * of the default extension (without the period).  Any or all of these
 * three parameters may be NULL, for default behaviour.
 *
 * <dwFlags> are used as follows:
 *   -- PFFN_OPENFILE: prompt the user for the name of a file to open
 *   -- PFFN_SAVEFILE: prompt the user for the name of a file to save to
 *   -- PFFN_SHOWDEFAULT: on entry, <achFileName> will contain the name
 *      of a file to display as the "default file name"
 *   -- PFFN_OVERWRITEPROMPT: prompt the user before accepting allowing them
 *      to choose an existing file
 *   -- PFFN_UPPERCASE: make <achFileName> uppercase before returning
 */
BOOL
PromptForFileName(hwndOwner, hInst, achFileName, cchFileName,
    szCaption, szFilter, szDefExt, dwFlags)
HWND        hwndOwner;        // window that will own dialog box
HANDLE        hInst;            // module that contains the resources
LPSTR        achFileName;        // where to put file name
WORD        cchFileName;        // size of <achFileName>
LPSTR       szCaption;        // caption string
LPSTR       szFilter;        // file filter
LPSTR       szDefExt;        // default extension
DWORD        dwFlags;        // random flags
{
    OPENFILENAME    ofname;        // parameter block
    BOOL        f;

    /* the initial file name is "" unless PFFN_SHOWDEFAULT is given */
    if (!(dwFlags & PFFN_SHOWDEFAULT))
        achFileName[0] = 0;

    /* fill in the other fields of <ofname> */
    ofname.lpstrFilter = szFilter;
    ofname.lpstrTitle = szCaption;
    ofname.lpstrDefExt = szDefExt;
    ofname.lStructSize = sizeof(ofname);
    ofname.hwndOwner = hwndOwner;
    ofname.hInstance = hInst;
    ofname.lpstrCustomFilter = NULL;
    ofname.nMaxCustFilter = 0;
    ofname.nFilterIndex = 1;
    ofname.lpstrFile = achFileName;
    ofname.nMaxFile = cchFileName;
    ofname.lpstrFileTitle = NULL;
    ofname.nMaxFileTitle = 0;
    ofname.lpstrInitialDir = NULL;
    ofname.Flags = OFN_HIDEREADONLY |
        (dwFlags & PFFN_OVERWRITEPROMPT ? OFN_OVERWRITEPROMPT : 0);
    ofname.lCustData = NULL;
    ofname.lpfnHook = NULL;
    ofname.lpTemplateName = NULL;

    /* prompt the user for the file name */
    if (dwFlags & PFFN_OPENFILE)
        f = GetOpenFileName(&ofname);
    else
        f = GetSaveFileName(&ofname);

    /* make the name uppercase if requested */
    if (f && (dwFlags & PFFN_UPPERCASE))
        AnsiUpper(achFileName);

    return f;
}
