///////////////////////////////////////////////////////////////////
// fmgr.h: A file manager class to handle file-based objects.
// NOTE: Fmgr objects are reference counted, and meant to be
// allocated dynamically only.
// Copyright(c) 1989-1993 Azarona Software. All rights reserved.
//////////////////////////////////////////////////////////////////
#ifndef H_FMGR
#define H_FMGR
#include <stdio.h>
#include "exchdlr.h"
#include "countobj.h"

struct FmgrHeader { // Goes at the beginning of every fmgr file
  long fs;    // Address to first block of "heap" free space
  long fe;    // Address of byte after end of file
  long hs;    // Address of the start of the "heap"
  char sg[4]; // Signature used for every fmgr-based file
};

struct FBlkHeader { // For free blocks
  int check_word;   // For file integrity checks. Optional.
  unsigned len;     // Length of block not including header.
  long next;        // Pointer to next free block
};

const long CURRADDR = -1; // Indicates current location

// File status bits used in Fmgr look like this
//
// 210
// |||---- good = 1, bad = 0
// |+----- open = 1, closed = 0
// +------ read/write = 1, read/only = 0

class Fmgr : public CountedObj {
public:
  enum Io_op      { fetch, store, seek };
  enum AccessMode { read_write, read_only };
  enum CheckWord  { FREEBLK = 0xFEFE };
protected:
  char name[128];    // Name of the file
  static char sg[4]; // Signature for fmgr style files
  FmgrHeader h;      // File header as stored in memory
  FILE *fp;          // Stream file handle
  Io_op lastop;      // Last I/O operation
  char status;       // See comments above
  void InitHdr();
  void ReadHdr();
  void WriteHdr();
  void FetchFBlkHdr(FBlkHeader &h, long p);
  void StoreFBlkHdr(const FBlkHeader &h, long p);
  long Reclaim(unsigned nbytes);
public:
  Fmgr();
  virtual ~Fmgr();
  virtual int Create(char *fname, long static_sz = 0);
  virtual int Open(char *fname, AccessMode mode = read_write);
  virtual void Close(int flush=1);
  virtual void Flush();
  long Alloc(unsigned nbytes);
  void Free(unsigned nbytes, long p);
  void Fetch(void *d, unsigned n, long p = CURRADDR);
  void Store(const void *d, unsigned n, long p = CURRADDR);
  void Seek(long ofs, int seek_mode = SEEK_SET);
  int IsOpen() const;
  int ReadOnly() const;
  int ReadyForWriting() const;
  void ClearErr();
  void Throw(ExceptionCode ec, char *src=0, int lno=0, char *msg=0);
  int OK() const;
  int operator!() const;
  operator const int () const;
};

// I don't usually do this, but I'm going to here

#include "fmgrptr.h"

inline void Fmgr::StoreFBlkHdr(const FBlkHeader &h, long p)
// Writes out free block header.
{
  Store(&h, sizeof(FBlkHeader), p);
}

inline int Fmgr::IsOpen() const
// Returns true open status bit is 1
{
  return (status & 0x02);
}

inline int Fmgr::ReadOnly() const
{
  return (status & 0x04) == 0;
}

inline void Fmgr::ClearErr()
{
  excp->Clear();
  status |= 0x01; // Set good bit
}

inline int Fmgr::OK() const
// We're okay if we're open and in a good state.
{
  return (status & 0x03) == 0x03;
}

inline int Fmgr::ReadyForWriting() const
// We're ready for writing if we're ok and not read-only
{
  return (status & 0x07) == 0x07;
}

inline int Fmgr::operator!() const
// We're not okay if not open or not in a good state.
{
  return (status & 0x03) != 0x03;
}

inline Fmgr::operator int() const
// Returns one if we're OK.
{
  return (status & 0x03) == 0x03;
}

#endif
