/**********************************************************************
** Program:    Memory Manager
** File:       memman.c
** Model:      medium
** Windows:    3.0
** Version:    1.0
** Author:     hb
** Date:       12.6.92
** Functions:  CreateHeap       creates a static heap. Only one allowed!
**             DestroyHeap      destroys the heap
**            *AllocMem         allocates a piece of memory in the heap
**            *ReAllocMem       tries to grow an allocated piece of memory
**             FreeMem          frees a piece of memory
** internal functions:
**             GetOffset        calculates the offset in the heap for a pointer
**            *GetPointer       calculates a pointer from a heapoffset
**             GetHeapIndex     returns the index of a heapnode in the heaplist
**             GetFreeIndex     returns the index of a freenode in the freenodelist
**             GetFreeHeapIndex returns the next unused index in the heaplist
**             GetFreeFreeIndex returns the next unused index in the freenodelist
**             MeltFreeList     tries to merge adjacent nodes in the freenodelist
**             SortFreeNodes    Quicksort for the freenodelist
***************************************************************************/
#pragma message  ("Modul: "__FILE__"  Version 1.0, Date:   "__TIMESTAMP__)
/******* Includes ********/
#include <windows.h>
#define MAINMEM
#include "dclmem.h"

/***CreateHeap*********************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   CreateHeap creates a heap
**             - does not create a heap in realmode
**             - creates only one heap because all handles stored in global vars.
**             - create a freenodelist and a heapnodelist with fixed number of entries
**             - heap, freenode and heapnodes are locked until destroyheap is called
** Return:     error
** Author:     hb
** Date:       12.6.92
****************************************************************************/
#pragma message   ("Compiling:         CreateHeap")
int CreateHeap (lSize, nMaxNodes)
long          lSize;
int           nMaxNodes;
{
   DWORD           dwInfo,
                   dwListSize;
   long            l;

/* is there already a heap? **/
   if (hHeap != NULL)
      return (IDE_ONLYONEHEAP);
/* get mode info **/
   dwInfo = GetWinFlags();
/* calculate segment boundaries **/
   if (dwInfo & WF_ENHANCED)
      lSegBorder = 0L;
   else if (dwInfo & WF_STANDARD)
      lSegBorder = 17L;
   else
      return (IDE_REALMODE);
/* try to allocate room for the heapnodes in global memory **/
   dwListSize = sizeof (struct node);
   dwListSize *= nMaxNodes;
   hHeapNodes = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, dwListSize);
   if (hHeapNodes == NULL)
      return (IDE_NOMEMAVAIL);
/* try to allocate room for the freenodes in global memory **/
   hFreeNodes = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, dwListSize);
   if (hFreeNodes == NULL)
   {
      GlobalFree (hHeapNodes);
      hHeapNodes = NULL;
      return (IDE_NOMEMAVAIL);
   }
/* try to allocate room for the heap in global memory **/
   hHeap = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, lSize);
   if (hHeap == NULL)
   {
      GlobalFree (hHeapNodes);
      GlobalFree (hFreeNodes);
      hHeapNodes = NULL;
      hFreeNodes = NULL;
      return (IDE_NOMEMAVAIL);
   }
/* lock all handles to get basepointers */
   HeapNodes = (LPNODE) GlobalLock (hHeapNodes);
   FreeNodes = (LPNODE) GlobalLock (hFreeNodes);
   if (HeapNodes == NULL || FreeNodes == NULL)
   {
      GlobalUnlock (hHeapNodes);
      GlobalUnlock (hFreeNodes);
      GlobalFree (hHeapNodes);
      GlobalFree (hFreeNodes);
      hHeapNodes = NULL;
      hFreeNodes = NULL;
      return (IDE_MEMERROR);
   }
/* calculate real size of heap **/
   lHeapSize = GlobalSize (hHeap);
   hpBase = (char huge *) GlobalLock (hHeap);
   if (hpBase == NULL)
   {
      GlobalFree (hHeap);
      hHeap = NULL;
      return (IDE_MEMERROR);
   }
/* create the startup entries in freenodes, one entry per segment **/
   for (l=0L,nFreeUsed=0;l<lHeapSize;nFreeUsed++,l+=SEGMENT)
   {
      FreeNodes[nFreeUsed].lOffset = (long) nFreeUsed * SEGMENT;
      if ((lHeapSize - l) / SEGMENT > 0L)
         FreeNodes[nFreeUsed].lSize = SEGMENT - lSegBorder;
      else
         FreeNodes[nFreeUsed].lSize = lHeapSize - ((long) nFreeUsed * SEGMENT);
   }
   nHeapUsed = 0;
   nNodes = nMaxNodes;
   return (IDE_OK);
}
/***DestroyHeap********************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   DestroyHeap destroys the heap
** Return:     error
** Author:     hb
** Date:       12.6.92
****************************************************************************/
#pragma message   ("Compiling:         DestroyHeap")
int DestroyHeap ()
{
   HANDLE       h;
   BOOL         b;

/* is ther a heap ? **/
   if (hHeap == NULL)
      return (IDE_NOHEAP);
/* free nodelists and heap */
   GlobalUnlock (hHeapNodes);
   GlobalUnlock (hFreeNodes);
   GlobalFree (hHeapNodes);
   GlobalFree (hFreeNodes);
   hHeapNodes = NULL;
   hFreeNodes = NULL;
   b = GlobalUnlock (hHeap);
   h = GlobalFree (hHeap);
   hHeap = NULL;
   hpBase = NULL;
   nFreeUsed = 0;
   nHeapUsed = 0;
   nNodes = 0;
   return (IDE_OK);
}
/***AllocMem***********************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   AllocMem dynamically gets a piece of memory from the heap
**             and returns a FAR pointer to it
** Return:     Pointer to object, NULL if error
** Author:     hb
** Date:       12.6.92
****************************************************************************/
#pragma message   ("Compiling:         AllocMem")
void far *AllocMem (lSize)
long lSize;
{
   int      i,
            nIndex,
            nError;
   void far *lpReturn;

/* Init **/
   lpReturn = NULL;
/* if size == 0 return error **/
   if (lSize == 0L)
      return (lpReturn);
/* do we have a free heapnode ? **/
   nIndex = GetFreeHeapIndex ();
/* search for block large enough for the request **/
   for (i=0;nIndex != nNodes && i<nFreeUsed;i++)
   {
      if (FreeNodes[i].lSize >= lSize)
      {
         lpReturn = GetPointer (FreeNodes[i].lOffset);
/******* create entry in heapnodes **/
         HeapNodes[nIndex].lSize = lSize;
         HeapNodes[nIndex].lOffset = FreeNodes[i].lOffset;
/******* split node in freenodes **/
         FreeNodes[i].lSize -= lSize;
         FreeNodes[i].lOffset += lSize;
         if (FreeNodes[i].lSize == 0L)
            FreeNodes[i].lOffset = 0L;
         break;
      }
   }
/* no room for the request, try to reorganize heap to satisfy the request **/
   if (lpReturn == NULL)
   {
      nError = MeltFreeList ();
/**** try again **/
      for (i=0;nError == IDE_OK && nIndex != nNodes && i<nFreeUsed;i++)
      {
         if (FreeNodes[i].lSize >= lSize)
         {
            lpReturn = GetPointer (FreeNodes[i].lOffset);
            HeapNodes[nIndex].lSize = lSize;
            HeapNodes[nIndex].lOffset = FreeNodes[i].lOffset;
            FreeNodes[i].lSize -= lSize;
            FreeNodes[i].lOffset += lSize;
            if (FreeNodes[i].lSize == 0L)
               FreeNodes[i].lOffset = 0L;
            break;
         }
      }
   }
   return (lpReturn);
}
/***ReAllocMem***********************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   ReAllocMem changes size of allocated previously allocated objct
**             - if size == 0, free node
**             - if there is no adjacent node to satisfy request, move block to 
**               new location
** Return:     Pointer to object, NULL if error
** Author:     hb
** Date:       12.6.92
****************************************************************************/
#pragma message   ("Compiling:         ReAllocMem")
void far *ReAllocMem (lp, lNewSize)
void far *lp;
long     lNewSize;
{
   long        lOffset,
               lOldSize,
               j;
   int         i,
               nIndex,
               nFreeIndex;
   LPSTR       lpNew, lpOld;

/* Init **/
   lpOld = lp;
   lOffset = GetOffset (lp);
   nIndex = GetHeapIndex (lOffset);
   lOldSize = HeapNodes[nIndex].lSize;
/* if size == 0 free node **/
   if (lNewSize == 0L)
   {
      FreeMem (lp);
      return (NULL);
   }
/* bad size */
   else if (lNewSize < 0L)
      return (NULL);

/* if node shrinks, add new memory to freenodes **/
   if (lNewSize < lOldSize)
   {
      HeapNodes[nIndex].lSize = lNewSize;
      nFreeIndex = GetFreeFreeIndex ();
      if (nFreeIndex != nNodes)
      {
         FreeNodes[nFreeIndex].lSize = lOldSize - lNewSize;
         FreeNodes[nFreeIndex].lOffset = lOffset + FreeNodes[nFreeIndex].lSize;
         if (nFreeUsed < nNodes - 1)
            nFreeUsed++;
      }
      return (lp);
   }
/* reorganize heap **/
   MeltFreeList ();
/* is there adjacent node with enough memory? **/
   lOffset += lOldSize;
   for (i=0;i<nFreeUsed;i++)
   {
/**** ok we got it **/
      if (FreeNodes[i].lOffset == lOffset)
      {
         FreeNodes[i].lSize -= lNewSize - lOldSize;
         FreeNodes[i].lOffset += lNewSize - lOldSize;
         if (FreeNodes[i].lSize <= 0L)
            FreeNodes[i].lOffset = 0L;
         HeapNodes[nIndex].lSize = lNewSize;
         return (lp);
      }
   }
/* we have to move the block **/
/* try to allocate the new block and copy the old one to new location **/
   if ((lpNew = (LPSTR) AllocMem (lNewSize)) == NULL)
      return (NULL);
   else
   {
      for (j=0L;j<HeapNodes[nIndex].lSize;j++)
         lpNew[j] = lpOld[j];
      FreeMem(lpOld);
      return (lpNew);
   }
   return (NULL);
}

/***FreeMem************************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   FreeMem frees a previously allocated block
** Return:     error
** Author:     hb
** Date:       12.6.92
****************************************************************************/
#pragma message   ("Compiling:         FreeMem")
int FreeMem (lp)
void far *lp;
{
   long          lOffset;
   int           nIndex,
                 nHeapIndex;

/* check for bad pointer **/
   if (lp == NULL || lp < hpBase)
      return (IDE_NULLPOINTER);
/* we need a new node in freenodes **/
   nIndex = GetFreeFreeIndex ();
/* calculate offset in heap **/
   lOffset = GetOffset (lp);
   nHeapIndex = GetHeapIndex (lOffset);
   if (nHeapIndex < IDE_OK)
      return (nHeapIndex);
/* add new node in freenodes **/
/* if there is no handle for that node the room ist lost forever **/
   if (nIndex < nNodes)
   {
      FreeNodes[nIndex].lSize = HeapNodes[nHeapIndex].lSize;
      FreeNodes[nIndex].lOffset = HeapNodes[nHeapIndex].lOffset;
   }
   HeapNodes[nHeapIndex].lSize = 0L;
   HeapNodes[nHeapIndex].lOffset = 0L;
   return (IDE_OK);
}
/***GetOffset**********************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   GetOffset calculates the offset inthe heap
** Return:     Offset in Heap, else 0L
** Author:     hb
** Date:       12.6.92
****************************************************************************/
#pragma message   ("Compiling:         GetOffset")
long GetOffset (lp)
void far *lp;
{
/* check for bad pointer **/
   if (lp == NULL || lp < hpBase)
      return (0L);
   return ((char huge *)lp - hpBase);
}
/***GetPointer*********************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   GetPointer
** Return:     Pointer in Heap
** Author:     hb
** Date:       12.6.92
****************************************************************************/
#pragma message   ("Compiling:         GetPointer")
void far *GetPointer (lOffset)
long lOffset;
{
   return (hpBase + lOffset);
}
/***GetHeapIndex*******************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   GetHeapIndex returns the node with specified heapoffset
** Return:     Index
** Author:     hb
** Date:       12.6.92
****************************************************************************/
#pragma message   ("Compiling:         GetHeapIndex")
int GetHeapIndex (lOffset)
long lOffset;
{
   int i;

   for (i=0;i<nHeapUsed;i++)
   {
      if (HeapNodes[i].lOffset == lOffset &&
          HeapNodes[i].lSize != 0L)
         return (i);
   }
   return (IDE_NOTINHEAP);
}
/***GetFreeIndex*******************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   GetFreeIndex 
** Return:     Index
** Author:     hb
** Date:       12.6.92
****************************************************************************/
#pragma message   ("Compiling:         GetFreeIndex")
int GetFreeIndex (lOffset)
long lOffset;
{
   int i;

   for (i=0;i<nFreeUsed;i++)
   {
      if (FreeNodes[i].lOffset == lOffset &&
          FreeNodes[i].lSize != 0L)
         return (i);
   }
   return (IDE_NOTINHEAP);
}
/***GetFreeHeapIndex*******************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   GetFreeHeapIndex returns the next unused node in heapnodes
** Return:     Index, else nNodes 
** Author:     hb
** Date:       12.6.92
****************************************************************************/
#pragma message   ("Compiling:         GetFreeHeapIndex")
int GetFreeHeapIndex ()
{
   int i,
       nReturn;

/* Init **/
   nReturn = nHeapUsed;
   if (nHeapUsed < nNodes - 1)
   {
      nHeapUsed++;
      return (nReturn);
   }
   for (i=0;i<nHeapUsed;i++)
   {
      if (HeapNodes[i].lSize == 0L)
         return (i);
   }
   return (nNodes);
}
/***GetFreeFreeIndex*******************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   GetFreeFreeIndex returns the next unused node in freenodes
** Return:     Index, else nNodes
** Author:     hb
** Date:       12.6.92
****************************************************************************/
#pragma message   ("Compiling:         GetFreeFreeIndex")
int GetFreeFreeIndex ()
{
   int i,
       nReturn,
       nError;

/* Init **/
   nReturn = nFreeUsed;
   if (nFreeUsed < nNodes - 1)
   {
      nFreeUsed++;
      return (nReturn);
   }
   for (i=0;i<nFreeUsed;i++)
   {
      if (FreeNodes[i].lSize == 0L)
         return (i);
   }
/* try to reorganize heap to get less used freenodes */
   nError = MeltFreeList ();
/* try again **/
   for (i=0;nError == IDE_OK && i<nFreeUsed;i++)
   {
      if (FreeNodes[i].lSize == 0L)
         return (i);
   }
/* Nichts gefunden liefere nNodes **/
   return (nNodes);
}
/***MeltFreeList***********************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   MeltFreeList reaorganizes fragmented heap
**             freenodes ae sorted, then adjacent nodes are merged
** Return:     nothing
** Author:     hb
** Date:       12.6.92
****************************************************************************/
#pragma message   ("Compiling:         MeltFreeList")
int MeltFreeList (void)
{
   long          lOffset;
   int           nNode,
                 i,
                 j,
                 nError;

/* Init **/
   nError = IDE_NOMELT;

/* quicksort freenodes **/
   SortFreeNodes (FreeNodes, 0, nFreeUsed-1);

/* look for adjacent nodes **/
   for (i=0;i<nFreeUsed;i++)
   {
/**** if there is a candidate merge it **/
/**** exclude segment boundaries **/
      for (nNode=i+1,j=0;nNode<nFreeUsed && FreeNodes[nNode].lSize>0;nNode++)
      {
         lOffset = FreeNodes[i].lOffset+FreeNodes[i].lSize;
         if (FreeNodes[nNode].lOffset % SEGMENT != 0L &&
             lOffset == FreeNodes[nNode].lOffset)
         {
            FreeNodes[i].lSize += FreeNodes[nNode].lSize;
            FreeNodes[nNode].lSize = 0;
            FreeNodes[nNode].lOffset = 0L;
            nError = IDE_OK;
            j++;
         }
      }
      i += j;
   }
   return (nError);
}
/***SortFreeNodes**********************************************************
** Program:    Memory Manager
** File:       Memman.c
** Funktion:   SortFreeNodes sorts freenode with a quicksort
** Return:     nothing
** Author:     hb
** Date:       12.6.92
***************************************************************************/
#pragma message   ("Compiling:         SortFreeNodes")
void SortFreeNodes (nodeList, nLeft, nRight)
LPNODE      nodeList;
int         nLeft,
            nRight;
{
   register int i,
                j;
   NODE         nodeX,
                nodeY;

   i = nLeft;
   j = nRight;
   nodeX.lSize = nodeList[(nLeft+nRight) / 2].lSize;
   nodeX.lOffset = nodeList[(nLeft+nRight) / 2].lOffset;

   do
   {
      while (nodeList[i].lOffset < nodeX.lOffset && i < nRight)
         i++;
      while (nodeX.lOffset < nodeList[j].lOffset && j > nLeft)
         j--;

      if (i <= j)
      {
         nodeY.lSize = nodeList[i].lSize;
         nodeY.lOffset = nodeList[i].lOffset;
         nodeList[i].lSize = nodeList[j].lSize;
         nodeList[i].lOffset = nodeList[j].lOffset;
         nodeList[j].lSize = nodeY.lSize;
         nodeList[j].lOffset = nodeY.lOffset;
         i++;
         j--;
      }
   } while (i <= j);
   if (nLeft < j)
      SortFreeNodes (nodeList, nLeft, j);
   if (i < nRight)
      SortFreeNodes (nodeList, i, nRight);

   return;
}
