// menu.c

#include "midisx.h" 
#include <memory.h>

// add MIDI devices to the window menu  
void CreateDeviceMenuPopups(
    HWND hWnd)
{    
    UINT u;
    char achDeviceName[40];
   
    // insert menu items for input devices                          
    for (u = 0; u < uMidiInDevices; u++)
    {   
        if (MidiGetDeviceName(hWnd, MIDI_IN, u, achDeviceName))
            // there was an error, so skip this menu item
            // and discontinue inserting menu items because they would
            // be out of sequence with the device ID values
            break;

        InsertMenu(hMenu, IDM_INPUT_ALWAYS_KEEP_OPEN, MF_BYCOMMAND,
            IDM_INPUT + u, achDeviceName);
    }
                            
    if (u)                            
        // we inserted input devices so insert a separator
        InsertMenu(hMenu, IDM_INPUT_ALWAYS_KEEP_OPEN, MF_SEPARATOR, 0, NULL);
           
     // insert menu items for output devices                        
    for (u = 0; u < uMidiOutDevices; u++)
    {
        if (MidiGetDeviceName(hWnd, MIDI_OUT, u, achDeviceName))
            // there was an error, so skip this menu item
            // and discontinue inserting menu items because they would
            // be out of sequence with the device ID values
            break;

        InsertMenu(hMenu, IDM_OUTPUT_ALWAYS_KEEP_OPEN, MF_BYCOMMAND,
            IDM_OUTPUT + u, achDeviceName);
    }
    
    if (u)                            
        // we inserted output devices so insert a separator
        InsertMenu(hMenu, IDM_OUTPUT_ALWAYS_KEEP_OPEN, MF_SEPARATOR, 0, NULL);
}
         
void InitializeMenu(
    HWND hWnd)         
{ 
    UINT u, uMenuStatus;
   
    // initialize input device menu items
                                   
    for (u = 0; u < uMidiInDevices; u++)
        CheckMenuItem(hMenu, IDM_INPUT + u, 
            MidiInInfo.uDeviceID == u ? MF_CHECKED : MF_UNCHECKED);            
     
    // if input device is busy, prevent user from changing it   
    uMenuStatus = MidiInInfo.uDeviceStatus == MIDI_DEVICE_BUSY ? MF_GRAYED : MF_ENABLED;     
    for (u = 0; u < uMidiInDevices; u++)
        EnableMenuItem(hMenu, IDM_INPUT + u, uMenuStatus);          
                                         
    CheckMenuItem(hMenu, IDM_INPUT_ALWAYS_KEEP_OPEN, 
        MidiInInfo.fAlwaysKeepOpen ? MF_CHECKED : MF_UNCHECKED);   
                 
    // initialize output device menu items
                     
    for (u = 0; u < uMidiOutDevices; u++)
        CheckMenuItem(hMenu, IDM_OUTPUT + u, 
            MidiOutInfo.uDeviceID == u ? MF_CHECKED : MF_UNCHECKED);            
                     
    // if output device is busy, prevent user from changing it   
    uMenuStatus = MidiOutInfo.uDeviceStatus == MIDI_DEVICE_BUSY ? MF_GRAYED : MF_ENABLED;     
    for (u = 0; u < uMidiOutDevices; u++)
        EnableMenuItem(hMenu, IDM_OUTPUT + u, uMenuStatus);          
                                                                                          
    CheckMenuItem(hMenu, IDM_OUTPUT_ALWAYS_KEEP_OPEN, 
        MidiOutInfo.fAlwaysKeepOpen ? MF_CHECKED : MF_UNCHECKED); 
}

BOOL ProcessMenuCommand(
    HWND hWnd,
    WORD wParam, 
    LPARAM lParam)
{   
    HINSTANCE hAppInst;
    FARPROC lpProc;   
    UINT uError = 0;

    switch (wParam)
    {         
        case IDM_FILE_SAVE:
            SaveInputBuffer(hWnd);
            break;
                
        case IDM_FILE_EXIT:
            PostMessage(hWnd, WM_CLOSE, 0, 0L);
            break;  
            
        case IDM_INPUT_ALWAYS_KEEP_OPEN:        
            MidiInInfo.fAlwaysKeepOpen ^= TRUE;        
            break;
      
        case IDM_OUTPUT_ALWAYS_KEEP_OPEN: 
            MidiOutInfo.fAlwaysKeepOpen ^= TRUE;       
            break;
     
        case IDM_SYSEX_COMMAND_STRING:
            // edit the system exclusive command string
            hAppInst = GetWindowWord(hWnd, GWW_HINSTANCE);
            lpProc = MakeProcInstance(CommandStringProc, hAppInst); 
            DialogBox(hAppInst, "CommandStringBox", hWnd, lpProc);
            FreeProcInstance(lpProc);        
            break;
 
        case IDM_SYSEX_INPUT_BUFFER:
            // edit the system exclusive input buffer size
            hAppInst = GetWindowWord(hWnd, GWW_HINSTANCE);
            lpProc = MakeProcInstance(InputBufferProc, hAppInst);               
            DialogBox(hAppInst, "InputBufferBox", hWnd, lpProc);
            FreeProcInstance(lpProc);        
            break;  
       
        case IDM_SEND:
            // prompt user to save input buffer if there is unsaved data
            if (QuerySaveChanges(hWnd) == IDCANCEL)
                break;
    
            // clear input buffer so there is no confusion over whether
            // data is from the current operation or a previous one
            _fmemset(lpInputBuffer, 0, cbInputBuffer);
            fUnsavedData = FALSE;
                            
            // assign buffer and prepare MIDIHDR for input device before
            // beginning 2-way system exclusive transfer to avoid overhead later
            uError = MidiAssignBuffer(hWnd, &MidiInInfo, lpInputBuffer, 
                cbInputBuffer, TRUE);                      
            if (uError)
                break;                                          
                                               
            // assign buffer and prepare MIDIHDR for output device 
            uError = MidiAssignBuffer(hWnd, &MidiOutInfo, abSysExCommand,
                cbSysExCommand, TRUE);                    
            if (uError)
            {            
                // unprepare MIDIHDR for input device since MIDIHDR
                // for output device had an error
                MidiAssignBuffer(hWnd, &MidiInInfo, NULL, 0L, FALSE);
                break;
            }
            
            // begin recording now to be ready for returned data                 
            MidiRecord(hWnd, &MidiInInfo, TRUE);                                        
            
            // initiate system exclusive communications               
            // transmit system exclusive command string
            uError = MidiSendLongMsg(hWnd, &MidiOutInfo, TRUE);            
            if (uError)
            {            
                // unprepare MIDIHDRs since there was an error
                MidiAssignBuffer(hWnd, &MidiInInfo, NULL, 0L, FALSE); 
                MidiAssignBuffer(hWnd, &MidiOutInfo, NULL, 0L, FALSE);
                break;
            }
           
            // system exclusive message is being sent
            // prevent user from re-entering this section                                      
            EnableMenuItem(hMenu, IDM_SEND, MF_GRAYED);
            EnableMenuItem(hMenu, IDM_ABORT, MF_ENABLED);            
            DrawMenuBar(hWnd);            

            break;            

        case IDM_ABORT:
        {           
            HDC hDC;               
            UINT uOutError = 0;
            
            // user wants to abort the operation, which is either still
            // transmitting or receiving
            
            uOutError = MidiSendLongMsg(hWnd, &MidiOutInfo, FALSE);                      
            if (uOutError)
                break;
                
            // unprepare the MIDIHDR (no harm if not actually in use)
            uOutError = MidiAssignBuffer(hWnd, &MidiOutInfo, NULL, 0L, FALSE);
                                     
            // stop recording even if there was an error with the output device
            uError = MidiRecord(hWnd, &MidiInInfo, FALSE);                
            if (uError)
                break;
            
            // unprepare the MIDIHDR
            uError = MidiAssignBuffer(hWnd, &MidiInInfo, NULL, 0L, FALSE);    

            if (uOutError || uError)
                break;
                       
            // restore application menu
            EnableMenuItem(hMenu, IDM_ABORT, MF_GRAYED);                
            EnableMenuItem(hMenu, IDM_SEND, MF_ENABLED); 
            DrawMenuBar(hWnd);
          
            // reinitialize the input buffer in case some data was received
            _fmemset(lpInputBuffer, 0, cbInputBuffer);
                
            // clear the display
            hDC = GetDC(hWnd);
            DisplayData(hDC);           // see display.c
            ReleaseDC(hWnd, hDC);           
            
            break;
        }
               
        case IDM_HELP_MIDISX_HELP:
        {   
            hAppInst = GetWindowWord(hWnd, GWW_HINSTANCE);
            lpProc = MakeProcInstance(MidiSXHelpProc, hAppInst);               
            DialogBox(hAppInst, "MidiSXHelpBox", hWnd, lpProc);
            FreeProcInstance(lpProc);
            break;
        }
                               
        case IDM_HELP_ABOUT:
        {   
            hAppInst = GetWindowWord(hWnd, GWW_HINSTANCE);
            lpProc = MakeProcInstance(AboutProc, hAppInst);               
            DialogBox(hAppInst, "AboutBox", hWnd, lpProc);
            FreeProcInstance(lpProc);
            break;
        }            
        
        default:
        {
            UINT uError;
            
            if (wParam >= IDM_INPUT && wParam < IDM_INPUT + uMidiInDevices)
            {
                // a MIDI input device was selected
                           
                if (wParam - IDM_INPUT == MidiInInfo.uDeviceID)                
                {   
                    // previously selected input device is being deselected
                    // close old input device (now none is selected) 
                    uError = MidiActivateDevice(hWnd, &MidiInInfo, 
                        MIDI_ABANDON);
                    if (uError)
                        // unable to close old input device (busy or error)
                        break;
                        
                    MidiInInfo.uDeviceID = NO_MIDI;                    
                }    
                else                
                {                    
                    // new device selected 
                    if (MidiInInfo.uDeviceID != NO_MIDI)
                    {                        
                        BOOL fAlwaysKeepOpen = MidiInInfo.fAlwaysKeepOpen;
                        
                        // clear old input device status and close  
                        MidiInInfo.fAlwaysKeepOpen = FALSE;                     
                        uError = MidiActivateDevice(hWnd, &MidiInInfo, 
                            MIDI_ABANDON);                
                                         
                        // restore status
                        MidiInInfo.fAlwaysKeepOpen = fAlwaysKeepOpen;                    
                        
                        if (uError)
                            // unable to close old input device (busy or error))
                            break;
                    }
                    
                    // open new input           
                    MidiInInfo.uDeviceID = wParam - IDM_INPUT;
                    MidiActivateDevice(hWnd, &MidiInInfo, MIDI_OPEN);                   
                }               
            }
            else if (wParam >= IDM_OUTPUT && 
                    wParam < IDM_OUTPUT + uMidiOutDevices)
            { 
                // a MIDI output device was selected
                     
                if (wParam - IDM_OUTPUT == MidiOutInfo.uDeviceID)                
                {
                     // previously selected output device is being deselected
                     // close old output device (now none is selected) 
                    uError = MidiActivateDevice(hWnd, 
                        &MidiOutInfo, MIDI_ABANDON);
                        
                    if (uError)
                        // unable to close old output device (busy or error)
                        break;                                                     
                        
                    MidiOutInfo.uDeviceID = NO_MIDI;                  
                }
                else                
                {
                    // new device selected 
                    if (MidiOutInfo.uDeviceID != NO_MIDI)
                    {                        
                        BOOL fAlwaysKeepOpen = MidiOutInfo.fAlwaysKeepOpen;
                        
                        // clear old output device status and close   
                        MidiOutInfo.fAlwaysKeepOpen = FALSE;                        
                        
                        uError = MidiActivateDevice(hWnd, 
                            &MidiOutInfo, MIDI_ABANDON);
                        
                        // restore status    
                        MidiOutInfo.fAlwaysKeepOpen = fAlwaysKeepOpen;
                        
                        if (uError)
                            // unable to close old output device (busy or error)
                            break;
                    }           

                    // open new output                    
                    MidiOutInfo.uDeviceID = wParam - IDM_OUTPUT;
                    MidiActivateDevice(hWnd, &MidiOutInfo, MIDI_OPEN);                          
                }              
            }
            else
                // we didn't handle this message
                return (FALSE);      
        }
    }
 
    // we handled the message   
    return (TRUE);
}
          