// midi.c

#define NOCOMM // suppress non-standard definitions which cause warnings

#ifndef _WINDOWS
#define _WINDOWS
#endif

#define WINVER 0x030a

#include <windows.h>
#include "midi.h"

// private declaration                    
void MidiShowError(HWND, UINT, UINT);
                    
/////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       MidiInit(LPMIDIINFO, LPMIDIINFO)
//  PURPOSE:        Determines number of MIDI devices in system and 
//                  initializes the caller's MIDI information blocks.
//  RETURN VALUE:   Returns TRUE if successful, FALSE if either or both
//                  requested MIDI device types are unavailable.
//                  The uDeviceType flag in MIDIINFO is set to NO_MIDI if
//                  the requested device is not available.
/////////////////////////////////////////////////////////////////////////////
BOOL MidiInit(
    LPMIDIINFO lpMidiInInfo,
    LPMIDIINFO lpMidiOutInfo)
{    
    // get the number of MIDI devices in the system
    uMidiInDevices = midiInGetNumDevs();
    uMidiOutDevices = midiOutGetNumDevs();

    // initialize the caller's MIDI information blocks
    // note: the application could specify NULL for one of the
    // information block addresses if it wasn't going to use it
    
    if (lpMidiInInfo)
    {        
        // initialize the MIDI input information block
        lpMidiInInfo->uDeviceType     = uMidiInDevices ? MIDI_IN : NO_MIDI;
        lpMidiInInfo->uDeviceID       = NO_MIDI;  
        lpMidiInInfo->hMidiIn         = NULL;     
        lpMidiInInfo->fAlwaysKeepOpen = FALSE;
        lpMidiInInfo->uDeviceStatus   = MIDI_DEVICE_IDLE;
    }
                      
    if (lpMidiOutInfo)
    { 
        // initialize the MIDI output information block 
        lpMidiOutInfo->uDeviceType     = uMidiOutDevices ? MIDI_OUT : NO_MIDI;
        lpMidiOutInfo->uDeviceID       = NO_MIDI;
        lpMidiOutInfo->hMidiOut        = NULL;
        lpMidiOutInfo->fAlwaysKeepOpen = FALSE;   
        lpMidiOutInfo->uDeviceStatus   = MIDI_DEVICE_IDLE;
    }

    if ((lpMidiInInfo && !uMidiInDevices)  || 
            (lpMidiOutInfo && !uMidiOutDevices))
    {   
        char achError[64];
        
        wsprintf((LPSTR)achError, (LPSTR)"No MIDI %s%s%sdevices are present",
            (LPSTR)(uMidiInDevices ? "" : "input "),
            (LPSTR)(uMidiInDevices || uMidiOutDevices ? "" : "or "),
            (LPSTR)(uMidiOutDevices ? "" : "output "));

        MessageBox(NULL, achError, "MidiInit", MB_ICONSTOP);
        
        // return failure
        return (FALSE);
    }
        
    // return success
    return (TRUE);
}

/////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       MidiGetDeviceName(HWND, UINT, UINT, LPSTR)
//  PURPOSE:        Get the name of the specified MIDI device
//  RETURN VALUE:   0 if successful, else MMSYSTEM error value > 0 or
//                  NO_MIDI if there is an application error.
///////////////////////////////////////////////////////////////////////////// 
UINT MidiGetDeviceName(
    HWND hWnd,              // caller's window handle
    UINT uDeviceType,       // MIDI_IN or MIDI_OUT
    UINT uDeviceID,         // MIDI device ID
    LPSTR lpszBuffer)       // points to buffer for storing device name
{
    UINT uError = 0;
    
    if (uDeviceType == MIDI_IN)
    {
        MIDIINCAPS mic;        
        
        uError = midiInGetDevCaps(uDeviceID, &mic, sizeof(mic));
        if (uError)
        {
            MidiShowError(hWnd, uError, uDeviceType);
            return (uError);
        }
        
        wsprintf(lpszBuffer, (LPSTR)(mic.szPname));        
    }
    else if (uDeviceType == MIDI_OUT)
    {
        MIDIOUTCAPS moc;
        
        uError = midiOutGetDevCaps(uDeviceID, &moc, sizeof(moc));
        if (uError)
        {
            MidiShowError(hWnd, uError, uDeviceType);
            return (uError);
        }
        
        wsprintf(lpszBuffer, (LPSTR)(moc.szPname));        
    }
    else
    {
        OutputDebugString("\r\nMidiInit: uDeviceType bad");    
        return (NO_MIDI);
    }
       
    // return success
    return (0);
}

/////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       MidiActivateDevice(HWND, LPMIDIINFO, UINT)
//  PURPOSE:        Open or close (activate or deactivate) the specified 
//                  MIDI device.
//  RETURN VALUE:   0 if successful activation/deactivation, else MMSYSTEM
//                  error value > 0, NO_MIDI if there is an application 
//                  error or MIDI_BUSY if a request to ABANDON_MIDI is not 
//                  completed.
/////////////////////////////////////////////////////////////////////////////
UINT MidiActivateDevice(
    HWND hWnd,              // caller's window handle
    LPMIDIINFO lpMidiInfo,  // address of MIDI information block   
    UINT uActivateMode)     // open device if == OPEN_MIDI
                            // close device if == CLOSE_MIDI unless 
                            // lpMidiInfo->fAlwaysKeepOpen == TRUE or
                            // lpMidiInfo->uDeviceStatus == MIDI_DEVICE_BUSY
                            // if == ABANDON_MIDI, close even if
                            // lpMidiInfo->fAlwaysKeepOpen == TRUE, but 
                            // prompt user first if lpMidiInfo->uDeviceStatus
                            // MIDI_DEVICE_BUSY
{
    UINT uError = 0;
    int nReturn;
   
    if (lpMidiInfo->uDeviceType == MIDI_OUT)
    {     
        if (uActivateMode == MIDI_OPEN)
        {   
            HMIDIOUT hMidiOut;
         
            // app is being activated or user has made request
               
            if (lpMidiInfo->hMidiOut)
                // already open so do nothing
                return (0);
            
            // open the output device
            uError = midiOutOpen(&hMidiOut, lpMidiInfo->uDeviceID, 
                hWnd, NULL, CALLBACK_WINDOW);
            if (uError)
                // couldn't open this device
                lpMidiInfo->uDeviceID = NO_MIDI;
            else
                // device was opened so remember the device handle
                lpMidiInfo->hMidiOut = hMidiOut;
        }
        else if (uActivateMode == MIDI_CLOSE)
        {   
            // app is being deactivated

            if (!lpMidiInfo->hMidiOut) 
                // already closed so do nothing
                return (0);
             
            if (lpMidiInfo->fAlwaysKeepOpen || 
                    lpMidiInfo->uDeviceStatus == MIDI_DEVICE_BUSY)
                // should remain open while app is inactive
                return (0);
   
            // close the output
            uError = midiOutClose(lpMidiInfo->hMidiOut);
            
            if (!uError)
                lpMidiInfo->hMidiOut = NULL;
        }
        else if (uActivateMode == MIDI_ABANDON)
        {
            if (!lpMidiInfo->hMidiOut) 
                // already closed so do nothing
                return (0);
 
            // user requests close
             
            if (lpMidiInfo->uDeviceStatus == MIDI_DEVICE_BUSY)
            {
                nReturn = MessageBox(hWnd, 
                    "MIDI output still in use - abandon it?",
                    "MidiActivateDevice", MB_YESNO);
                    
                if (nReturn == IDNO)
                    return (MIDI_BUSY);
                    
                // abandon current operation; allow device to be closed            
                uError = midiOutReset(lpMidiInfo->hMidiOut);                             
                               
                if (!uError)                         
                {
                    // clear the status
                    lpMidiInfo->uDeviceStatus = MIDI_DEVICE_IDLE;
                    
                    // unprepare the MIDIHDR (no harm if not actually in use)
                    uError = MidiAssignBuffer(hWnd, lpMidiInfo, NULL, 0L, 
                        FALSE);                        
                }
            }
                
            if (!uError)
                // close the output
                uError = midiOutClose(lpMidiInfo->hMidiOut);
                
            if (!uError)
                lpMidiInfo->hMidiOut = NULL;
        }
        else
            // application error
            return (NO_MIDI);
    }
    else if (lpMidiInfo->uDeviceType == MIDI_IN)
    {    
        if (uActivateMode == MIDI_OPEN)
        {
            HMIDIIN hMidiIn;
            
            // app is being activated or user has made request
            
            if (lpMidiInfo->hMidiIn)
                // already open so do nothing
                return (0);
                                     
            // open the input device
            uError = midiInOpen(&hMidiIn, lpMidiInfo->uDeviceID, 
                hWnd, NULL, CALLBACK_WINDOW);
            if (uError)
                // couldn't open this device
                lpMidiInfo->uDeviceID = NO_MIDI;
            else
                // device was opened so remember the device handle             
                lpMidiInfo->hMidiIn = hMidiIn; 
        }
        else if (uActivateMode == MIDI_CLOSE)
        {
            // app is being deactivated
            
            if (!lpMidiInfo->hMidiIn) 
                // already closed so do nothing
                return (0);
 
            if (lpMidiInfo->fAlwaysKeepOpen || 
                    lpMidiInfo->uDeviceStatus == MIDI_DEVICE_BUSY)
                // should remain open while app is inactive
                return (0);
   
             // close the input
            uError = midiInClose(lpMidiInfo->hMidiIn); 
            
            if (!uError)
                lpMidiInfo->hMidiIn = NULL;  
        }
        else if (uActivateMode == MIDI_ABANDON)        
        {   
            if (!lpMidiInfo->hMidiIn) 
                // already closed so do nothing
                return (0);        
        
            // user requests close
             
            if (lpMidiInfo->uDeviceStatus == MIDI_DEVICE_BUSY)
            {
                nReturn = MessageBox(hWnd, 
                    "MIDI input still in use - abandon it?",
                    "MidiActivateDevice", MB_YESNO);
                    
                if (nReturn == IDNO)
                    return (MIDI_BUSY); 
              
                // abandon current operation; allow device to be closed            
                uError = midiInReset(lpMidiInfo->hMidiIn);       
                         
                if (!uError)     
                {
                    // clear the status
                    lpMidiInfo->uDeviceStatus == MIDI_DEVICE_IDLE;                                        
                    
                    // unprepare the MIDIHDR
                    uError = MidiAssignBuffer(hWnd, lpMidiInfo, NULL, 0L, 
                        FALSE);                       
                }
            }
                
            if (!uError)
                // close the input     
                uError = midiInClose(lpMidiInfo->hMidiIn);
                
            if (!uError)                
                lpMidiInfo->hMidiIn = NULL;
        }
        else
            // application error
            return (NO_MIDI);
    }
    else
    {
        // application error        
        OutputDebugString("\r\nMidiActivateDevice: uDeviceType bad");    
        return (NO_MIDI);
    }    

    if (uError)    
        MidiShowError(hWnd, uError, lpMidiInfo->uDeviceType);
        
    return (uError);   
}                                        

/////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   MidiSendShortMsg(HWND, LPMIDIINFO, BYTE, BYTE, BYTE, BYTE)
//  PURPOSE:    Send status, channel and data.
//  RETURNS:    0 on success, else MMSYSTEM error value > 0 or NO_MIDI if
//              there is an application error.
/////////////////////////////////////////////////////////////////////////////
UINT MidiSendShortMsg(
    HWND hWnd,          // caller's window handle
    LPMIDIINFO lpMidiOutInfo, // address of MIDI output info block
    BYTE cStatus,       // MIDI status byte for this message
    BYTE cChannel,      // MIDI channel byte for this message
    BYTE cData1,        // MIDI data byte for this message
    BYTE cData2)        // 2nd MIDI data byte for this message (may be 0)
{   
    DWORD dwMsg;
    UINT uError = 0;
    
    if (lpMidiOutInfo->uDeviceType != MIDI_OUT)
    {
        MessageBox(hWnd, "Must use a MIDI output device",
            "MidiSendShortMsg", MB_ICONSTOP);
        return (NO_MIDI);
    }
    
    if (!lpMidiOutInfo->hMidiOut)   
    {
        MessageBox(hWnd, "Must open MIDI output device first",
            "MidiSendShortMsg", MB_ICONSTOP);
        return (NO_MIDI);
    }
    
    // pack the message and send it
    dwMsg = MAKEMIDISHORTMSG(cStatus, cChannel, cData1, cData2);
    uError = midiOutShortMsg(lpMidiOutInfo->hMidiOut, dwMsg); 
     
    if (uError)  
        MidiShowError(hWnd, uError, MIDI_OUT);
        
    return (uError);  
}

/////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   MidiAssignBuffer(HWND, LPMIDIINFO, BOOL)
//  PURPOSE:    Assign lpData and dwBufferLength members and prepare the
//              MIDIHDR.  Also add the buffer if it is an input buffer.
//              If the third parameter is false, unprepare and reinitialize
//              the header.
//  RETURNS:    0 on success, else MMSYSTEM error value > 0 or NO_MIDI if
//              there is an application error.
/////////////////////////////////////////////////////////////////////////////
UINT MidiAssignBuffer(
    HWND hWnd,                // caller's window handle
    LPMIDIINFO lpMidiInfo,    // address of MIDI info block 
    LPSTR lpData,             // address of buffer, NULL if cleaning up
    DWORD dwBufferLength,     // size of buffer in bytes, 0L if cleaning up
    BOOL fAssign)             // TRUE = assign, FALSE = cleanup
{   
    UINT uError = 0;  
    
    if (lpMidiInfo->uDeviceType == MIDI_IN)
    {     
        if (!lpMidiInfo->hMidiIn)   
        {   
            if (!fAssign && lpMidiInfo->uDeviceStatus == MIDI_DEVICE_ABANDONED)
                // don't return an error for this case because if the user aborts
                // the record process after the header has been prepared but
                // before midiInStart is called, the input buffer will still
                // be returned to the application and an MM_MIM_LONGDATA message 
                // will also be received, just as it would when recording
                // successfully completes.  So it is reasonable for this function
                // to be called (with fAssign = FALSE) when hMidiIn = 0, as long
                // as uDeviceStatus = MIDI_DEVICE_ABANDONED.
                return (0);
            else
            {            
                // all other cases are an application error          
                MessageBox(hWnd, "Must open MIDI input device first",
                    "MidiAssignBuffer", MB_ICONSTOP);       
                    
                // because this failed call might result in an input or output
                // device being reset (if the application is written to do so),
                // an MM_MOM_DONE or MM_MIM_LONGDATA message could be sent to 
                // the application.  This might result in an additional call to
                // this routine, so set the device status to prevent another 
                // error message
                lpMidiInfo->uDeviceStatus = MIDI_DEVICE_ABANDONED;
                
                return (NO_MIDI);      
            }
        }
    }    
    else if (lpMidiInfo->uDeviceType == MIDI_OUT)
    {
        if (!lpMidiInfo->hMidiOut)   
        {                           
            if (!fAssign && lpMidiInfo->uDeviceStatus == MIDI_DEVICE_ABANDONED)
            {
                // clear the device status            
                lpMidiInfo->uDeviceStatus = MIDI_DEVICE_IDLE; 
               
                // don't return an error for this case because if the user aborts
                // the transmission of a long message before it completes the
                // buffers will be marked as done and returned to the application, 
                // just as when recording successfully completes.  So it is ok
                // for this function to be called (with fAssign = FALSE) when
                // hMidiIn = 0, as long as uDeviceStatus = MIDI_DEVICE_ABANDONED.
                return (0);
            }    
            else
            {
                // all other cases are an application error        
                MessageBox(hWnd, "Must open MIDI output device first",
                    "MidiAssignBuffer", MB_ICONSTOP);       
                    
                // because this failed call might result in an input or output
                // device being reset (if the application is written to do so),
                // an MM_MOM_DONE or MM_MIM_LONGDATA message could be sent to 
                // the application.  This might result in an additional call to
                // this routine, so set the device status to prevent another 
                // error message
                lpMidiInfo->uDeviceStatus = MIDI_DEVICE_ABANDONED;
                                    
                return (NO_MIDI);
            }
        }
    }
    else
    {
        OutputDebugString("\r\nMidiAssignBuffer: uDeviceType bad");    
        return (NO_MIDI);
    }
        
    if (fAssign)
    {   
        // check for the buffer's address and size
        if (!lpData || !dwBufferLength)   
        {
            MessageBox(hWnd, "Must specify a buffer and size",
                "MidiAssignBuffer", MB_ICONSTOP);
            return (NO_MIDI);
        }
       
        // assign buffer to the MIDIHDR 
        lpMidiInfo->MidiHdr.lpData = lpData;
        lpMidiInfo->MidiHdr.dwBufferLength = dwBufferLength;
        
        // prepare the MIDIHDR        
        if (lpMidiInfo->uDeviceType == MIDI_OUT)            
            uError = midiOutPrepareHeader(lpMidiInfo->hMidiOut, 
                &(lpMidiInfo->MidiHdr), sizeof(MIDIHDR));
        else
        {   
            // lpMidiInfo->uDeviceType == MIDI_IN         
            
            uError = midiInPrepareHeader(lpMidiInfo->hMidiIn, 
                &(lpMidiInfo->MidiHdr), sizeof(MIDIHDR));
          
            if (!uError)
                // add the input buffer associated with the MIDIHDR
                uError = midiInAddBuffer(lpMidiInfo->hMidiIn, 
                    &(lpMidiInfo->MidiHdr), sizeof(MIDIHDR));                 
        }    
    }
    else
    {   
        // unprepare the MIDIHDR    
        
        if (lpMidiInfo->uDeviceType == MIDI_OUT) 
        {
            if (lpMidiInfo->MidiHdr.dwFlags != MHDR_DONE)                
                uError = midiOutReset(lpMidiInfo->hMidiOut);
                
            if (!uError)
                uError = midiOutUnprepareHeader(lpMidiInfo->hMidiOut, 
                    &(lpMidiInfo->MidiHdr), sizeof(MIDIHDR));  
        }                          
        else // lpMidiInfo->uDeviceType == MIDI_IN     
        {             
            if (lpMidiInfo->MidiHdr.dwFlags != MHDR_DONE)                
                uError = midiInReset(lpMidiInfo->hMidiIn);
                
            if (!uError)           
                uError = midiInUnprepareHeader(lpMidiInfo->hMidiIn, 
                    &(lpMidiInfo->MidiHdr), sizeof(MIDIHDR));         
        }                        
        
        if (!uError)
        {
            // reinitialize MIDIHDR to guard against casual re-use
            lpMidiInfo->MidiHdr.lpData = NULL;
            lpMidiInfo->MidiHdr.dwBufferLength = 0; 
            
            // clear the device status 
            lpMidiInfo->uDeviceStatus = MIDI_DEVICE_IDLE;          
        }
    }   
    
    if (uError) 
        MidiShowError(hWnd, uError, lpMidiInfo->uDeviceType);
        
    return (uError); 
}
                                                                     
/////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   MidiSendLongMsg(HWND, LPMIDIINFO, BOOL)
//  PURPOSE:    Send system exclusive message or series of short messages.
//              Abort the process if the third parameter is FALSE.
//  RETURNS:    0 on success, else MMSYSTEM error value > 0 or NO_MIDI if
//              there is an application error.
/////////////////////////////////////////////////////////////////////////////
UINT MidiSendLongMsg(
    HWND hWnd,                // caller's window handle
    LPMIDIINFO lpMidiOutInfo, // address of MIDI output info block    
    BOOL fSend)               // send if TRUE, abort if FALSE and busy
{   
    UINT uError = 0;
    
    if (lpMidiOutInfo->uDeviceType != MIDI_OUT)
    {
       MessageBox(hWnd, "Must use a MIDI output device",
            "MidiSendLongMsg", MB_ICONSTOP);
        return (NO_MIDI);
    }
    
    if (!lpMidiOutInfo->hMidiOut)   
    {
        if (!fSend && lpMidiOutInfo->uDeviceStatus == MIDI_DEVICE_ABANDONED)
            // don't return an error for this case because if the user aborts
            // the transmission of a long message before it completes the
            // buffers will be marked as done and returned to the application, 
            // just as when recording successfully completes.  So it is ok
            // for this function to be called (with fSend = FALSE) when
            // hMidiIn = 0, as long as uDeviceStatus = MIDI_DEVICE_ABANDONED.
            return (0);
        else
        {            
            // all other cases are an application error
            MessageBox(hWnd, "Must open MIDI output device first",
                "MidiSendLongMsg", MB_ICONSTOP);
            return (NO_MIDI);
        }
    }
    
    if (fSend)
    {               
        // send the long message
        uError = midiOutLongMsg(lpMidiOutInfo->hMidiOut,        
            &(lpMidiOutInfo->MidiHdr), sizeof(MIDIHDR));
        if (!uError)
            lpMidiOutInfo->uDeviceStatus = MIDI_DEVICE_BUSY;
    }
    else if (lpMidiOutInfo->MidiHdr.dwFlags != MHDR_DONE)
    {
        // abort the current message
        uError = midiOutReset(lpMidiOutInfo->hMidiOut);
        if (!uError)
            // set the device status because buffer(s) have been marked as
            // done and returned to the application
            lpMidiOutInfo->uDeviceStatus = MIDI_DEVICE_ABANDONED;
    }
    else
        // tried to abort but the operation was already complete
        lpMidiOutInfo->uDeviceStatus = MIDI_DEVICE_IDLE;
                                                  
    if (uError)    
        MidiShowError(hWnd, uError, MIDI_OUT);
        
    return (uError);   
}
    
/////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   MidiRecord(HWND, LPMIDIINFO, BOOL)
//  PURPOSE:    Start or stop recording MIDI data.
//  RETURNS:    0 on success, else MMSYSTEM error value > 0 or NO_MIDI if
//              there is an application error.
/////////////////////////////////////////////////////////////////////////////
UINT MidiRecord(
    HWND hWnd,                  // caller's window handle
    LPMIDIINFO lpMidiInInfo,    // address of MIDI input info block
    BOOL fRecord)               // start recording if TRUE, stop if FALSE
{   
    UINT uError = 0; 
       
    if (lpMidiInInfo->uDeviceType != MIDI_IN)
    {
        MessageBox(hWnd, "Must use a MIDI input device",
            "MidiRecord", MB_ICONSTOP);
        return (NO_MIDI);
    }
    
    if (!lpMidiInInfo->hMidiIn)   
    {
    
        if (!fRecord && lpMidiInInfo->uDeviceStatus == MIDI_DEVICE_ABANDONED)
            // don't return an error for this case because if the user aborts
            // the record process after the header has been prepared but
            // before midiInStart is called, the input buffer will still
            // be returned to the application and an MM_MIM_LONGDATA message 
            // will also be received, just as it would when recording
            // successfully completes.  So it is reasonable for this function
            // to be called (with fRecord = FALSE) when hMidiIn = 0, as long
            // as uDeviceStatus = MIDI_DEVICE_ABANDONED.
            return (0);
        else
        {    
            // all other cases are an application error
            MessageBox(hWnd, "Must open MIDI input device first",
                "MidiRecord", MB_ICONSTOP);
            
            return (NO_MIDI);
        }
    }

    if (fRecord)
    {           
        // begin recording       
        uError = midiInStart(lpMidiInInfo->hMidiIn);
        if (!uError)
            lpMidiInInfo->uDeviceStatus = MIDI_DEVICE_BUSY;
    }        
    else
    {                  
        // stop recording
        // note - don't bother checking to see if the device is idle
        // because midiInStop has no effect in that case
        uError = midiInStop(lpMidiInInfo->hMidiIn);
        if (!uError)
            lpMidiInInfo->uDeviceStatus = MIDI_DEVICE_IDLE;
    }
                  
    if (uError)    
        MidiShowError(hWnd, uError, MIDI_IN);  
        
    return (uError);   
}
  
/////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       MidiExit(HWND)
//  PURPOSE:        Closes MIDI devices unless they are busy and user
//                  wants to keep them open.
//  RETURN VALUE:   TRUE if OK to exit, else FALSE.
/////////////////////////////////////////////////////////////////////////////
BOOL MidiExit(
    HWND hWnd,                  // caller's window handle
    LPMIDIINFO lpMidiInInfo,    // address of MIDI input info block
    LPMIDIINFO lpMidiOutInfo)   // address of MIDI output info block
{    
    UINT uResult = 0;  
 
    // close MIDI devices if necessary
    
    if (lpMidiInInfo && lpMidiInInfo->hMidiIn)
    {     
        uResult = MidiActivateDevice(hWnd, lpMidiInInfo, MIDI_ABANDON);        
        if (uResult == MIDI_BUSY)
            // input device still in use, so fail
            return (FALSE); 
            
        // other than MIDI_BUSY, we ignore errors so app can close
    }
        
    if (lpMidiOutInfo && lpMidiOutInfo->hMidiOut)
    {   
        uResult = MidiActivateDevice(hWnd, lpMidiOutInfo, MIDI_ABANDON);
        if (uResult == MIDI_BUSY)
            // output device still in use, so fail
            return (FALSE);
            
        // other than MIDI_BUSY, we ignore errors so app can close
    }
        
    // notify caller it's ok to terminate
    return (TRUE);
}
        
/////////////////////////////////////////////////////////////////////////////
//  FUNCTION:       MidiShowError(HWND, UINT, UINT)
//  PURPOSE:        Displays an error from MMSYSTEM.
//  RETURN VALUE:   void
/////////////////////////////////////////////////////////////////////////////
void MidiShowError(
    HWND hWnd,          // caller's window handle
    UINT uError,        // MMSYSTEM error value
    UINT uDeviceType)   // MIDI_IN or MIDI_OUT
{
    char achTitle[20], achBuf[MIDI_ERRMSG_SIZE];
    UINT uGetError = 0;

    switch (uDeviceType)
    {
        case MIDI_IN:
            uGetError = midiInGetErrorText(uError, achBuf, MIDI_ERRMSG_SIZE);
            if (uGetError)
                // uError was an invalid parameter
                midiInGetErrorText(uGetError, achBuf, MIDI_ERRMSG_SIZE);
            wsprintf((LPSTR)achTitle, (LPSTR)"MIDI Input Error");
            break;

        case MIDI_OUT:
            uGetError = midiOutGetErrorText(uError, achBuf, MIDI_ERRMSG_SIZE);
            if (uGetError)
                // uError was an invalid parameter
                midiOutGetErrorText(uGetError, achBuf, MIDI_ERRMSG_SIZE);              
            wsprintf((LPSTR)achTitle, (LPSTR)"MIDI Output Error");
            break;

        default:
            OutputDebugString("MidiShowError: Bad MIDI device type\r\n");
            return;
    }

    MessageBox(hWnd, achBuf, achTitle, MB_ICONSTOP | MB_TASKMODAL);
}

