// midisx.c
//
// Written by Microsoft Product Support Services, Windows Developer Support
// Copyright (c) 1994 Microsoft Corporation. All rights reserved.
//
// 04/21/94     first version by Allan Tamm
//
// MidiSX demonstrates how to send a system exclusive message and receive system
// exclusive data returned by a MIDI device.  Wrapper functions are provided 
// which handle errors and simplify the application code.  Manifest constants 
// describing the MIDI message protocol are also provided, as well as macros for 
// cracking MIDI messages.  This application opens any MIDI devices it has
// selected when it receives activation and releases them when it loses
// activation unless it is in the middle of a MIDI operation at that time.

#include "midisx.h"

char * szAppName = "MidiSX";
char * szAppClassName = "MidiSX";

int PASCAL WinMain(
    HINSTANCE hInstance,
    HINSTANCE hPrevInstance,
    LPSTR lpCmdLine,
    int nCmdShow)
{   
    HWND hWnd;
    MSG msg;
 
    // look for existence of previous instance     
    hWnd = FindWindow(szAppClassName, NULL);
    if (hWnd)
    {
        // application is already running so tell user
        if (IsIconic(hWnd))
            ShowWindow(hWnd, SW_NORMAL);
        else            
            BringWindowToTop(hWnd);
            
        MessageBox(hWnd, "Only a single instance of this application can run",
            szAppName, MB_OK);
            
        return (FALSE);
    }
    
    // perform initializations
    if (!InitApplication(hInstance))
        return (FALSE);

    while (GetMessage(&msg, NULL, NULL, NULL))
    {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    return (msg.wParam);
}

BOOL InitApplication(
    HINSTANCE hInstance)
{
    WNDCLASS wc; 
    HWND hWnd;  
                     
    wc.style         = NULL;
    wc.lpfnWndProc   = MainWndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = LoadIcon(hInstance, MAKEINTRESOURCE(ID_MIDISX_ICO));
    wc.hbrBackground = GetStockObject(WHITE_BRUSH); 
    wc.lpszClassName = szAppClassName;
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);    
    wc.lpszMenuName  = "MidiSXMenu";

    if (!RegisterClass(&wc))
        return (FALSE);

    // create this instance's main window
    hWnd = CreateWindow(szAppClassName, szAppName, WS_OVERLAPPEDWINDOW,
        CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, 
        NULL, NULL, hInstance, NULL);

    if (!hWnd)
        return (FALSE);
                            
    // perform MIDI device initialization functions                            
    if (!MidiInit(&MidiInInfo, &MidiOutInfo))       
        return (FALSE);
        
    // add MIDI devices to the window menu
    CreateDeviceMenuPopups(hWnd);       
    
    // select default input and output ports
    MidiInInfo.uDeviceID = 0;
    MidiOutInfo.uDeviceID = 0;                   

    // initialize ASCII command string for edit control
    wsprintf((LPSTR)achCommandString, (LPSTR)"%s", (LPSTR)"F0 00 F7");
   
    // initialize corresponding binary string used in system exclusive    
    TranslateSysExString();   
 
    // set MIDI input buffer size
    cbInputBuffer = INPUT_BUFFER_SIZE; 
    
    // allocate MIDI input buffer
    if (!AllocInputBuffer(hWnd, cbInputBuffer))
        return (FALSE);
 
    // display the window
    ShowWindow(hWnd, SW_NORMAL);
    UpdateWindow(hWnd);
                 
    return (TRUE);
}

long FAR PASCAL __export MainWndProc(
    HWND hWnd,
    UINT uMessage,
    WORD wParam,
    LONG lParam)
{
    HDC hDC;
    
    switch (uMessage)
    {     
        case WM_SIZE:                      
            // set variables for data display                        
            SizeDisplay(hWnd, LOWORD(lParam), HIWORD(lParam)); // see display.c
  
            break;
 
        // case MM_MOM_DONE:                
            // This is the callback message sent by mmsystem upon the successful
            // completion of midiOutLongMessage.  It can be ignored here since
            // MIDI recording was initiated *before* midiOutLongMessage was
            // called, ensuring the MIDI input device will be ready to receive
            // incoming system exclusive data.
     
        case MM_MIM_LONGDATA:
        {
            UINT uInError = 0;
            UINT uOutError = 0;    
             
            // this callback message sent by mmsystem indicates that the input
            // buffer was returned, so stop recording
            uInError = MidiRecord(hWnd, &MidiInInfo, FALSE);     
                           
            if (!uInError)
                // unprepare and reinitialize the input MIDIHDR
                uInError = MidiAssignBuffer(hWnd, &MidiInInfo, NULL, 0L, FALSE);
                         
            // don't forget to unprepare and reinitialize the output MIDIHDR 
            uOutError = MidiAssignBuffer(hWnd, &MidiOutInfo, NULL, 0L, FALSE);           
                          
            if (uInError || uOutError)
                break;
           
            // restore application menu
            EnableMenuItem(hMenu, IDM_ABORT, MF_GRAYED);                
            EnableMenuItem(hMenu, IDM_SEND, MF_ENABLED); 
            DrawMenuBar(hWnd);
             
            // check first byte in input buffer for system exclusive status byte
            if (*lpInputBuffer == MIDI_SYSEX)
                // set flag indicating input buffer contains a sysex message
                fUnsavedData = TRUE;
       
            // display contents of input buffer
            hDC = GetDC(hWnd);
            DisplayData(hDC);           // see display.c
            ReleaseDC(hWnd, hDC);      
            
            break;
        }
         
        case WM_PAINT:
        {
            PAINTSTRUCT ps;
                          
            // display contents of input buffer
            hDC = BeginPaint(hWnd, &ps);       
            DisplayData(hDC);            // see display.c   
            EndPaint(hWnd, &ps); 
            
            break;
        }
        
        case WM_ACTIVATEAPP:   // application is gaining or losing activation
        {           
            if (MidiInInfo.uDeviceID != NO_MIDI)
                // open or close current input device per app activation
                MidiActivateDevice(hWnd, &MidiInInfo, 
                    wParam ? MIDI_OPEN : MIDI_CLOSE);               
                
            if (MidiOutInfo.uDeviceID != NO_MIDI)
                // open or close current output device per app activation
                MidiActivateDevice(hWnd, &MidiOutInfo, 
                    wParam ? MIDI_OPEN : MIDI_CLOSE);
          
            break;
        }
            
        case WM_INITMENU:
            InitializeMenu(hWnd);       // see menu.c
            break;
        
        case WM_COMMAND:
            if (ProcessMenuCommand(hWnd, wParam, lParam)) // see menu.c
                // we processed the message so break
                break;
            else 
                // we didn't process the message - let Windows do it
                return (DefWindowProc(hWnd, uMessage, wParam, lParam));
                       
        case WM_CREATE:    
            // remember application menu handle
            hMenu = GetMenu(hWnd);
            
            // disable the Abort menu
            EnableMenuItem(hMenu, IDM_ABORT, MF_GRAYED);     

            // establish screen metrics
            GetCharSize(hWnd);              // see display.c                           
 
            break;
            
        case WM_CLOSE:
        case WM_QUERYENDSESSION:        
            if (QuerySaveChanges(hWnd) == IDCANCEL) 
                // user decided not to exit
                break;
                
            if (!MidiExit(hWnd, &MidiInInfo, &MidiOutInfo))
                // user does not wish to end long MIDI operation
                break;
                    
            DestroyWindow(hWnd);
            
            if (uMessage == WM_QUERYENDSESSION)
                // continue shutdown
                return (1L);
                
            break;
     
        case WM_DESTROY:   
            // clean up    
            ReleaseInputBuffer();                
           
            PostQuitMessage(0);
            break;

        default:
            return (DefWindowProc(hWnd, uMessage, wParam, lParam));
    }

    return (0L);
}    
 
