/*
	-------- MIKEE -------
	Mouse Interface w/ Keyboard Events Emulation
	
	by Jerzy R. Chrzaszcz, ICS WUT, Warsaw, Poland


	Interfacing Microsoft Mouse driver to the keyboard-driven application
	program by the means of brute force: direct filling of keyboard buffer.
	There is no dedicated mouse cursor on the screen, so the application
	should handle and display its own text cursor as usually.

	Assumptions:
		1. the mouse is used in text mode only,
		2. the mouse driver has been installed before,
		3. the code is to be compiled in large memory model,

	Interface:
		int		mouse_open(void)
		void		mouse_close(void)
		void		mouse_off(void)			
		void		mouse_on(void)	   	
		void		mouse_move_off(void)	
		void		mouse_move_on(void)	
		void		mouse_keys_off(void)	
		void		mouse_keys_on(void)	
		void		mouse_fast(void)		
		void		mouse_slow(void)		

	Mouse movement translates to arrows; button codes and relative speed
	may be changed. Initially left key is ENTER and right key is ESCAPE.
	There are some provisions made for the middle key, but it is ignored
	in this version. The activity of the whole mouse or only the mouse keys
	may be suspended. The movement detection may be suspended as well.
*/

#include		<stdlib.h>
#include		"mikee.h"



unsigned		mouse_mask = MOUSE_MASK;
unsigned		mouse_left = ENTER;
unsigned		mouse_right = ESC;
char			mouse_lastx = 0;
char			mouse_lasty = 0;
int			mouse_stepx = MOUSE_STEPX;
int			mouse_stepy = MOUSE_STEPY;


int	mouse_open (void)
/*
	starting cooperation with msmouse
		returns 0 if initialization fails,
		returns -1 otherwise
*/
{
	static char		(*stack)[1024];
	void				(*fun)();

	if (!msm_init ())
		return(0);
	msm_setareax (0,MOUSE_XMAX);
	msm_setareay (0,MOUSE_YMAX);
	msm_setcurpos (MOUSE_XMAX>>1,MOUSE_YMAX>>1);
	fun = mouse_isr;
	msm_signal (mouse_mask,fun,stack);
	return (-1);
}


void	mouse_close (void)
/*
	closing cooperation with msmouse
*/
{
	msm_term ();
}


void	mouse_isr (unsigned event,unsigned state,unsigned curposx, unsigned curposy)
/*
	installable mouse interrupt service routine
	watching for clicks and tracking the movement
*/
{
	static int	prvposx = MOUSE_XMAX>>1,
					prvposy = MOUSE_YMAX>>1;
	int			difx, dify;
	char			moved = 0;


	if ((event &= mouse_mask) == 0)
		return;
	if (event & MSM_MOVED)
	{
		difx = curposx - prvposx;
		dify = curposy - prvposy;
		if (abs (difx) > mouse_stepx)
		{
			if (difx > 0)
			{
				putkey(RT_ARROW);
				prvposx += mouse_stepx;
			}
			else
			{
				putkey(LT_ARROW);
				prvposx -= mouse_stepx;
			}
			moved = 1;
		}
		if ((abs (dify) > mouse_stepy) && !moved)
		{
			if (dify > 0)
			{
				putkey(DN_ARROW);
				prvposy += mouse_stepy;
			}
			else
			{
				putkey(UP_ARROW);
				prvposy -= mouse_stepy;
			}
		}
		if ((curposx == 0) || (curposx == MOUSE_XMAX))
		{
			prvposx = curposx = MOUSE_XMAX>>1;
			msm_setcurpos (curposx,curposy);
		}
		if ((curposy == 0) || (curposy == MOUSE_YMAX))
		{
			prvposy = curposy = MOUSE_YMAX>>1;
			msm_setcurpos (curposx,curposy);
		}
	}
	else
	{
		if ((event & MSM_L_PRESSED) && (state == LEFT_BUTTON))
			putkey(mouse_left);
		else
		if ((event & MSM_R_PRESSED) && (state == RIGHT_BUTTON))
			putkey(mouse_right);
	}
}

