/////////////////////////////////////////////////////////////
// mnodeptr.h: Cached-Mnode pointer class definition.
// *** This file is based on the Bucket<>, Coptr<>,
// and Cache<> templates, as given in the book, "Practical
// Data Structures in C++" by Bryan Flamig.
// Copyright(c) 1993 Azarona Software. All rights reserved.
/////////////////////////////////////////////////////////////
#ifndef H_MNODEPTR
#define H_MNODEPTR
#include "bucketb.h"
#include "coptrb.h"
#include "cacheb.h"
#include "placenew.h"
#include "mnode.h"

class MnodeBucket : public Bucketb, public Mnode {
public:
  MnodeBucket();
  virtual void Fetch(Fmgr &f);
  virtual void Store(Fmgr &f);
};

inline MnodeBucket::MnodeBucket()
// Calls default Bucketb() and Mnode() constructors
{
}

inline void MnodeBucket::Fetch(Fmgr &f)
// Fetch object data from the file f
{
  f.Fetch((Mnode *)this, sizeof(Mnode), addr);
  dirty = 0;
}

inline void MnodeBucket::Store(Fmgr &f)
// Stores object data into the file f
{
  f.Store((Mnode *)this, sizeof(Mnode), addr);
  dirty = 0;
}


class MnodeCache : public Cacheb {
protected:
  MnodeBucket *buckets;
public:
  MnodeCache(int n);
  virtual ~MnodeCache();
};

inline MnodeCache::MnodeCache(int n)
// Constructor to create n buckets. Note that Bucket's 
// default constructor is called.
: Cacheb(buckets = new MnodeBucket[n], n, sizeof(MnodeBucket))
{
}

inline MnodeCache::~MnodeCache()
// This probably shouldn't be inline, but we put it here
// for convenience.
{
  if (fptr) Flush();
  Disconnect();
  delete[] buckets;
}


class MnodePtr : public Coptrb {
public:
  friend void *operator new(size_t n, MnodePtr &cp, long p);
  void Delete();
  MnodePtr(MnodeCache &c, long p = 0);
  MnodePtr(const MnodePtr &c);
  void operator=(const MnodePtr &c);
  void operator=(long p);
  MnodeBucket &operator*();
  MnodeBucket *operator->();
};

inline void *operator new(size_t n, MnodePtr &cp, long p=0)
// Allocates a new object to reside at address p in the file.
// If p is non-zero, the object is presumed to already
// allocated. The allocation takes place by calling Alloc(),
// which not only allocates the object on disk, but also
// reserves a cache bucket for it. A Bucketb pointer to this
// bucket is returned by Alloc(), which we typecast as a
// MnodeBucket pointer. Then, we do a clever typecast
// to a Mnode pointer, which gives us a pointer to the bucket's
// data. This pointer is returned, and is passed on to the assoc-
// iated Mnode constructor as the 'this' pointer. (We had to pass
// the pointer back as a void pointer to satisfy the syntax for
// the new operator.)
{
  return (Mnode *)((MnodeBucket *)(cp.Alloc(n, p)));
}

inline void MnodePtr::Delete()
// Frees the data associated with the Coptr c. The Coptr is
// then set to null.
{
  Coptrb::Delete(sizeof(Mnode));
}

inline MnodePtr::MnodePtr(MnodeCache &c, long p)
// General coptr constructor. Note that a cache
// is required. The coptr stays unbound till needed.
: Coptrb(c, p)
{
}

inline MnodePtr::MnodePtr(const MnodePtr &c)
// Coptr copy constructor. Note that destination coptr
// won't be bound until needed, even if source is bound.
: Coptrb(c)
{
}

inline void MnodePtr::operator=(const MnodePtr &c)
{
  Copy(c);
}

inline void MnodePtr::operator=(long p)
{
  Release();
  addr = p;
}

inline MnodeBucket &MnodePtr::operator*()
// Access to the referenced bucket. If the pointer
// isn't bound to a bucket one is grabbed for it.
{
  if (!bound) Bind();
  return *(MnodeBucket *)bkt;
}

inline MnodeBucket *MnodePtr::operator->()
// Access to the referenced bucket. If the pointer
// isn't bound to a bucket, one is grabbed for it.
{
  if (!bound) Bind();
  return (MnodeBucket *)bkt;
}

#endif
