/****************************************************************************
 * Copyright (c) 1993, BSDI.  All rights reserved.
 * Library Name...  Object Engine V1.01
 * File Name......  objeng.hpp
 * Comments.......  This file contains the class definitions and #defines
 ***************************************************************************/
#ifndef _OEDBH_
#define _OEDBH_

#include <string.h>
#include <time.h>
#include <pxengine.h>

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#ifndef NULL
#define NULL 0
#endif

const int MAXFIELDS=64;
const int OE_MAXTABLES=32;
const int NOTINIT = 0;
const int FAIL = 1;
const int READY = 2;
const int TBLDIRBUFLEN = 80;
const int CASEINSENSITIVE = 1;
const int CASESENSITIVE = 0;

//-----------------------------------------------------------
typedef struct {
  char TableName[12];
  TABLEHANDLE tHdl;
  RECORDHANDLE rHdl;
  FIELDHANDLE indexID;
} OEHandle;


//----------------------------------------------------------------
typedef struct {
  char *data;
  unsigned size;
} BlobData;


//----------------------------------------------------------------
class Blob {
  BLOBHANDLE bHdl;
  RECORDHANDLE rHdl;
  FIELDHANDLE fHdl;
  int status;
public:
  Blob(OEHandle &, FIELDHANDLE);
  Blob &operator=(const Blob &inblobobj);

  // Inline wrapper functions.
  int Clone() { return(status = PXBlobClone(rHdl, fHdl)); }
  int Close(int accept) { return(status = PXBlobClose(bHdl, accept)); }
  int Drop() { return(status = PXBlobDrop(rHdl, fHdl)); }
  int OpenRead() { return(status = PXBlobOpenRead(rHdl, fHdl, &bHdl)); }
  int GetSize(unsigned long *size) { return(status = PXBlobGetSize(bHdl, size)); }
  int Get(unsigned size, unsigned long offset, void* buff) { return(status = PXBlobGet(bHdl, size, offset, buff)); }
  int QuickGet(int bufSize, void* dest, int *bytesRead) { return(status = PXBlobQuickGet(rHdl, fHdl, bufSize, dest, bytesRead)); }
  int OpenWrite(unsigned long size, int saveCurrent) { return(status = PXBlobOpenWrite(rHdl, fHdl, &bHdl, size, saveCurrent)); }
  int Put(unsigned size, unsigned long offset, void* buff) { return (status = PXBlobPut(bHdl, size, offset, buff)); }

  // These two functions allow you to simplify blob programming.
  // They work for buffers < 64K.
  int OEPut(char *buff, unsigned size);
  int OEGet(BlobData &bdata);
  int GetStatus() const { return status; }
  int GetRHdl() const { return rHdl; }
  int GetFHdl() const { return fHdl; }
};


//-----------------------------------------------------------
class ObjEng {
private:
  int status;
  int OESaveChanges;
  char filedir[TBLDIRBUFLEN];
  OEHandle RegisteredTables[OE_MAXTABLES];
  TABLEHANDLE tHdl;
  RECORDHANDLE rHdl;
  int TableInit();
public:
  ObjEng(char *tableDirectory, char *userName, char *netNamePath, int netCode);
  ObjEng();  //Uses all defaults.
  ~ObjEng();
  int GetIndex(long &index);
  int EnsureExist(char *tblname, int nFields, char **FieldNames, char **FieldTypes);
  int RegisterCompKey(char *tblname, int nIdxFields, FIELDHANDLE *IdxFieldHandles, char *keyName, int idxType, int caseMode);
  int RegisterKey(char *tblname, int nIdxFields, FIELDHANDLE *IdxFieldHandles, int idxType);
  int DropKey(char *tblname, FIELDHANDLE fldID);
  int Register(OEHandle &oehandle);
  int IsRegistered(OEHandle &oehandle);
  const char *GetFileDir() const { return filedir; }
  int GetStatus() const { return status; }
};

                                   
//-------------------------------------------------------------------------
class PersistClass {
protected:
  // Holds the name of and Paradox Engine handles for the table that
  // matches this persistant class.
  OEHandle oehandle;

  // The objectID is at the center of the persistance system.
  // It is the "link" between the object and a db table record.
  // It is assigned when a virgin object is Stored and cleared when 
  // Destroyed.  Each of the link functions resets objectID to that 
  // of the db record being retrieved when the client object is filled from 
  // the database.  The assignment operator does NOT copy it but the
  // copy constructor DOES copy it.
  long objectID;

  // "C" Types - used during the "LinkToField" function.
  char localtypes[MAXFIELDS]; 

  // Status holds the most recent Engine return code.  Most functions
  // return this value for user to test against.  Uses engine error
  // codes exclusively!
  int status;

  // Retrieve is defined by derived classes to indicate how the linking
  // required by the functions below is actually implemented.  It 
  // must be in the derived classes since they are the ones to know the
  // structure of the object/db record.
  virtual int Retrieve() = 0;
  // Utility to copy PersistClass data for the Copy constructor in the descendants
  void Copy(PersistClass &pc);
public:
  // Default class constructor.
  PersistClass(char *name, int indexID);
  // Used in the copy constructor.
  PersistClass() {}
  int SynchronizeDB();

  // Store places the contents of an object into the data table.  If
  // the object is not linked (no objectID) then an objectID is assigned.
  // Otherwise, the contents of the object are placed in the record
  // corresponding to the objectID.
  virtual int Store() = 0;

  // Table/C++ Class "Linking" functions - allow class instances to
  // be associated with DB records.
  virtual int LinkToKey(int, int);
  virtual int LinkToSKey(char *, int);
  int LinkToRecord(int position);
  int LinkToField(char *fieldname, void *value, int mode);
  int LinkToID(const long &ID);
  int LinkToFirst();
  int LinkToLast();
  int LinkToNext();
  int LinkToPrev();

  // "Unlinking" functions.  Also, the copy constructor and assignment 
  // operator will unlink the recipient of the copy/assignment.
  void Unlink();
  virtual int Destroy();

  // Utilities
  int GetFieldHandle(char *fldName, FIELDHANDLE &fieldHandle) { return(status = PXFldHandle(oehandle.tHdl, fldName, &fieldHandle)); }
  int NRecs(RECORDNUMBER & nrecs);
  int RecNum(RECORDNUMBER & recnum);
  int OEPutDate(const int &fld, const char *indate);
  int OEGetDate(const int &fld, char *indate);
  int OEPutDate(const int &, const struct tm &);
  int OEGetDate(const int &, struct tm &);
  // Network locking/unlocking/info/refresh functions.
  int LockFile(int lockType);
  int UnlockFile(int lockType);
  int LockRecord(LOCKHANDLE &lckHandle);
  int IsRecordLocked(int &isLocked);
  int UnlockRecord(int lckHandle);
  int GotoLock(LOCKHANDLE lckHandle);
  int LockTable(int lockType);
  int UnlockTable(int lockType);
  int IsTblChanged(int &isChanged);
  int TableRefresh();

  int GetStatus() const { return status; }
  TABLEHANDLE GetTblHdl() const { return oehandle.tHdl; }
  RECORDHANDLE GetRecHdl() const { return oehandle.rHdl; }
  const char *GetTblName() const { return oehandle.TableName; }
  long GetObjectID() const { return objectID; }
};

#endif
