

/* Recursive descent parser for integer expressions. */

#include "setjmp.h"
#include "math.h"
#include "ctype.h"
#include "stdlib.h"
#include "stdio.h"
#include "string.h"

enum tok_types {DELIMITER, VARIABLE, NUMBER, COMMAND,
                STRING, QUOTE};

enum tokens {PRINT=1, INPUT, IF, THEN, FOR, NEXT, TO,
             GOTO, GOSUB, RETURN, EOL, FINISHED, END};

extern char *prog;  /* points into the program */
extern char *p_buf; /* points to start of program */
extern jmp_buf e_buf; /* holds environment for longjmp() */
extern int variables[26]; /* variables */
extern struct commands { 
  char command[20];
  char tok;
} table[];

extern char token[80]; /* holds string representation of token */
extern char token_type; /* contains type of token */
extern char tok; /* holds the internal representation of token */

void eval_exp1(int *answer), eval_exp2(int *answer);
void eval_exp3(int *answer), eval_exp4(int *answer);
void eval_exp5(int *answer), eval_exp6(int *answer);
void atom(int *answer), putback(void);
void serror(int error);
int get_token(void), look_up(char *s);
int isdelim(char c), iswhite(char c);
int find_var(char *s);

/* These are the constants used to call serror() when
   a syntax error occurs.  Add more if you like.
   NOTE: SYNTAX is a generic error message used when
   nothing else seems appropriate.
*/
enum error_msg 
     {SYNTAX, UNBAL_PARENS, NO_EXP, EQUALS_EXP,
      NOT_VAR, LAB_TAB_FULL, DUP_LAB, UNDEF_LAB,
      THEN_EXP, TO_EXP, TOO_MNY_FOR, NEXT_WO_FOR,
      TOO_MNY_GOSUB, RET_WO_GOSUB, MISS_QUOTE};

/* Entry point into parser. */
void eval_exp1(int *answer)
{
  get_token(); 
  if(!*token) {
    serror(NO_EXP); 
    return; 
  }
  eval_exp2(answer); 
  putback(); /* return last token read to input stream */
}

/*  Add or subtract two terms. */
void eval_exp2(int *answer)
{
  register char  op; 
  int temp; 

  eval_exp3(answer); 
  while((op = *token) == '+' || op == '-') {
    get_token(); 
    eval_exp3(&temp); 
    switch(op) {
      case '-' :
        *answer = *answer - temp;
        break;
      case '+':
        *answer = *answer + temp;
        break;
    }
  }
}

/* Multiply or divide two factors. */
void eval_exp3(int *answer)
{
  register char  op; 
  int temp; 

  eval_exp4(answer); 
  while((op = *token) == '*' || op == '/') {
    get_token(); 
    eval_exp4(&temp); 
    switch(op) {
      case '*' :
        *answer = *answer * temp;
        break;
      case '/':
        *answer = *answer / temp;
        break;
    }
  }
}

/* Process integer exponent. */
void eval_exp4(int *answer)
{
  int temp, i, t; 

  eval_exp5(answer); 
  if(*token== '^') {
    get_token(); 
    eval_exp4(&temp); 
    if(temp==0) {
      *answer = 1;
      return;
    }
    i = *answer;
    for(t=temp-1; t>0;  t--) *answer = (*answer) * i;
  }
}

/* Is a unary + or -. */
void eval_exp5(int *answer)
{
  register char  op; 

  op = 0; 
  if((token_type==DELIMITER) && *token=='+' || *token=='-') {
    op = *token; 
    get_token(); 
  }
  eval_exp6(answer); 
  if(op=='-') *answer = -(*answer);
}

/* Process parenthesized expression. */
void eval_exp6(int *answer)
{
  if(*token == '(') {
    get_token(); 
    eval_exp2(answer); 
    if(*token != ')')
      serror(UNBAL_PARENS); 
    get_token(); 
  }
  else
    atom(answer); 
}

/* Find value of number or variable. */
void atom(int *answer)
{
  switch(token_type) {
    case VARIABLE:
      *answer = find_var(token); 
      get_token(); 
      return; 
    case NUMBER:
      *answer = atoi(token); 
      get_token();   
      return; 
    default:
      serror(SYNTAX); 
    }
}

/* Find the value of a variable. */
int find_var(char *s)
{
  if(!isalpha(*s)){
    serror(NOT_VAR); /* not a variable */
    return 0; 
  }
  return variables[toupper(*token)-'A']; 
}

/* Display an error message. */
void serror(int error)
{
  char *p, *temp;
  int linecount = 0;
  register int i;

  static char *e[]= {   
    "syntax error", 
    "unbalanced parentheses", 
    "no expression present",
    "equals sign expected",
    "not a variable",
    "Label table full",
    "duplicate label",
    "undefined label",
    "THEN expected",
    "TO expected",
    "too many nested FOR loops",
    "NEXT without FOR",
    "too many nested GOSUBs",
    "RETURN without GOSUB",
    "double qoutes needed"
  }; 
  printf("%s", e[error]); 

  p = p_buf;
  while(p != prog) {  /* find line number of error */
    p++;
    if(*p == '\r') {
      linecount++;
    }
  }
  printf(" in line %d\n", linecount);

  temp = p;  /* display line with error */
  for(i=0; i<20 && p>p_buf && *p!='\n'; i++, p--);
  for(; p<=temp; p++) printf("%c", *p);
    
  longjmp(e_buf, 1); /* return to save point */
}

/* Get a token. */
get_token(void)
{
  register char *temp;

  token_type = 0; 
  tok = 0;
  temp = token;

  if(*prog=='\0') { /* end of file */
    *token = 0;
    tok = FINISHED;
    return(token_type=DELIMITER);
  }

  while(iswhite(*prog)) ++prog;  /* skip over white space */

  if(*prog=='\r') { /* crlf */
    ++prog; ++prog;
    tok = EOL; *token='\r';
    token[1]='\n'; token[2]=0;
    return (token_type = DELIMITER);
  }

  if(strchr("+-*^/=;(),><", *prog)){ /* delimiter */
    *temp = *prog;
    prog++; /* advance to next position */
    temp++;
    *temp = 0; 
    return (token_type=DELIMITER);
  }
    
  if(*prog=='"') { /* quoted string */
    prog++;
    while(*prog!='"'&& *prog!='\r') *temp++ = *prog++;
    if(*prog=='\r') serror(MISS_QUOTE);
    prog++;*temp = 0;
    return(token_type=QUOTE);
  }
  
  if(isdigit(*prog)) { /* number */
    while(!isdelim(*prog)) *temp++ = *prog++;
    *temp = '\0';
    return(token_type = NUMBER);
  }

  if(isalpha(*prog)) { /* var or command */
    while(!isdelim(*prog)) *temp++ = *prog++;
    token_type = STRING;
  }
  
  *temp = '\0';

  /* see if a string is a command or a variable */
  if(token_type==STRING) {
    tok = look_up(token); /* convert to internal rep */
    if(!tok) token_type = VARIABLE;
    else token_type = COMMAND; /* is a command */
  }
  return token_type;
}

/* Return a token to input stream. */
void putback(void) 
{

  char *t; 

  t = token; 
  for(; *t; t++) prog--; 
}

/* Look up a token's internal representation in the
   token table.
*/
look_up(char *s)
{
  register int i;
  char *p;

  /* convert to lowercase */
  p = s;
  while(*p){ 
    *p = tolower(*p);
    p++;
  }

  /* see if token is in table */
  for(i=0; *table[i].command; i++)
      if(!strcmp(table[i].command, s)) return table[i].tok;
  return 0; /* unknown command */
}

/* Return true if c is a delimiter. */
isdelim(char c)
{
  if(strchr(" ;,+-<>/*%^=()", c) || c==9 || c=='\r' || c==0) 
    return 1;  
  return 0; 
}

/* Return 1 if c is space or tab. */
iswhite(char c)
{
  if(c==' ' || c=='\t') return 1;
  else return 0;
}


