
/* An example of object rotation. */

#include "dos.h"
#include "stdio.h"
#include "math.h"
#include "conio.h"
#include "ctype.h"

#define XMAX 319
#define YMAX 199

void line(int startx, int starty, int endx, int endy, 
         int color);
void mempoint(int x, int y, int color);
void palette(int pnum);
void mode(int mode_code);
void rotate_point(double theta, double *x, double *y,
                  int x_org, int y_org);
void display_object(double ob[][5], int lines);
void rotate_object(double ob[][5], double theta,
                   int x, int y, int lines);

/* house array */
double house[][5] = {
/* startx, starty, endx, endy, color */
  120, 120, 200, 120, 2, /* house */
  200, 120, 200, 80, 2,
  120, 80, 200, 80, 2,
  120, 80, 120, 120, 2,
  160, 60, 120, 80, 2, /* roof */
  160, 60, 200, 80, 2,
  155, 120, 155, 100, 2, /* door */
  155, 100, 165, 100, 2,
  165, 100, 165, 120, 2,
  130, 90, 130, 100, 2, /* windows */
  130, 90, 140, 90, 2,
  130, 100, 140, 100, 2,
  140, 90, 140, 100, 2,
  180, 90, 180, 100, 2,
  180, 90, 190, 90, 2,
  180, 100, 190, 100, 2,
  190, 90, 190, 100, 2
};

main(void)
{
  mode(4);  /* mode 4 graphics, palette 0 */
  palette(0);
  
  /* draw a box around the house to give perspective */
  line(70, 30, 260, 30, 1);
  line(70, 160, 260, 160, 1);
  line(70, 30, 70, 160, 1);
  line(260, 30, 260, 160, 1); 

  display_object(house, 17);
  rotate_object(house, 0.025, 160, 90, 17);
  mode(3);
  
  return 0;
}

/* Set the palette. */
void palette(int pnum)
{
  union REGS r;

  r.h.bh = 1; /* code for palette select */
  r.h.bl = pnum;
  r.h.ah = 11; /* set palette function */
  int86(0x10, &r, &r);
}

/* Set the video mode. */
void mode(int mode_code)
{
  union REGS r;

  r.h.al = mode_code;
  r.h.ah = 0;
  int86(0x10, &r, &r);
}

/* Draw a line in the specified color
   using Bresenham's integer based algorithm.
*/
void line(int startx, int starty, /* beginning of line */
         int endx, int endy, /* end of line */
         int color) /* color */
{
  register int t, distance;
  int x=0, y=0, delta_x, delta_y;
  int incx, incy;

  /* Compute the distances in both directions. */
  delta_x = endx-startx;
  delta_y = endy-starty;

  /* Compute the direction of the increment.
     An increment of 0 means either a vertical
     or horizontal line.
  */
  if(delta_x>0) incx = 1;
  else if(delta_x==0) incx = 0;
  else incx = -1;

  if(delta_y>0) incy = 1;
  else if(delta_y==0) incy = 0;
  else incy = -1;

  /* determine which distance is greater */
  delta_x = abs(delta_x);
  delta_y = abs(delta_y);
  if(delta_x>delta_y) distance = delta_x;
  else distance = delta_y;

  /* draw the line */
  for(t=0; t<=distance+1; t++) {
    mempoint(startx, starty, color);
    x += delta_x;
    y += delta_y;  
    if(x>distance) {
      x -= distance;
      startx += incx;
    }
    if(y>distance) {
      y -= distance;
      starty += incy;
    }
  }
}

/* Write a point directly to mode 4 graphics RAM. */
void mempoint(int x, int y, /* coordinate of point */
              int color) /* color */
{
  union mask {
    char c[2];
    int i;
  } bit_mask;
  int index, bit_position;
  unsigned char t;
  char xor; /* xor color in or overwrite */

  /* pointer to mode 4 graphics RAM */
  char far *ptr = (char far *) 0xB8000000; 

  bit_mask.i = 0xFF3F;  /* 11111111 00111111 in binary */

  /* check range for mode 4 */
  if(x<0 || x>XMAX || y<0 || y>YMAX) return;

  xor = color & 128; /* see if xor mode is set */
  color = color & 127; /* mask off high bit */
 
  /* set bit_mask and color bits to the right location */
  bit_position = x%4;
  color <<= 2*(3-bit_position);
  bit_mask.i >>= 2*bit_position;


  /* find the correct byte in screen memory */
  index = y*40 + (x >> 2);
  if(y % 2) index += 8152; /* if odd use 2nd bank */

  /* write the color */
  if(!xor) { /* overwrite mode */
    t = *(ptr+index) & bit_mask.c[0];
    *(ptr+index) = t | color;
  }
  else { /* xor mode */
    t = *(ptr+index) | (char) 0;
    *(ptr+index) = t ^ color;
  }
}

/* Rotate a point around the origin, specified by
   x_org and y_org, by angle theta. 
*/
void rotate_point(double theta, /* angle of rotation */
                  double *x, double *y, /* point to rotate */
                  int x_org, int y_org) /* origin */
{
  double tx, ty;

  /* normalize x and y */
  tx = *x - x_org;  
  ty = *y - y_org;
 
  /* rotate */
  *x = tx * cos(theta) - ty * sin(theta);
  *y = tx * sin(theta) + ty * cos(theta);

  /* return to PC coordinate values */
  *x += x_org;
  *y += y_org;

}

/* Rotate the specified object. */
void rotate_object(double ob[][5], /* object definition */
            double theta, /* angle of rotation in radians */
            int x, int y, /* location of origin */
            int lines) /* number of lines in image */
{
  register int j;
  char ch;

  for(;;) {
    ch = getch(); /* see which direction to rotate */
    switch(tolower(ch)) {
      case 'l':  /* counterclockwise */
        theta = theta < 0 ? -theta : theta;
        break;
      case 'r':  /* clockwise */
        theta = theta > 0 ? -theta : theta;
        break;
      default: return;
    }

    for(j=0; j<lines; j++) {     
      /* erase old line */
      line((int) ob[j][0], (int) ob[j][1],
        (int) ob[j][2], (int) ob[j][3], 0);
 
      rotate_point(theta, &ob[j][0], 
        &ob[j][1], x, y);

      rotate_point(theta, &ob[j][2], 
        &ob[j][3], x, y);

      line((int) ob[j][0], (int) ob[j][1],
        (int) ob[j][2], (int) ob[j][3], (int) ob[j][4]);
    }
  }
}

/* Display an object. */
void display_object(double ob[][5], int lines)
{
  register int i;

  for(i=0; i<lines; i++)
    line((int) ob[i][0], (int) ob[i][1],
      (int) ob[i][2], (int) ob[i][3], (int) ob[i][4]);
}


