

/* A Small BASIC interpreter */

#include "stdio.h"
#include "setjmp.h"
#include "math.h"
#include "ctype.h"
#include "stdlib.h"
#include "string.h"

#define NUM_LAB 100
#define LAB_LEN 10 
#define FOR_NEST 25
#define SUB_NEST 25
#define PROG_SIZE 10000

enum tok_types {DELIMITER, VARIABLE, NUMBER, COMMAND,
                STRING, QUOTE};

enum tokens {PRINT=1, INPUT, IF, THEN, FOR, NEXT, TO,
             GOTO, GOSUB, RETURN, EOL, FINISHED, END};

char *prog; /* points into the program */
char *p_buf; /* points to start of program */
jmp_buf e_buf; /* holds environment for longjmp() */

int variables[26]= {    /* 26 user variables,  A-Z */
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0
};

/* keyword lookup table */
struct commands { 
  char command[20]; /* string form */
  char tok; /* internal representation */
} table[] = { /* Commands must be entered lowercase */
  "print", PRINT, /* in this table. */
  "input", INPUT,
  "if", IF,
  "then", THEN,
  "goto", GOTO,
  "for", FOR,
  "next", NEXT,
  "to", TO,
  "gosub", GOSUB,
  "return", RETURN,
  "end", END,
  "", END  /* mark end of table */
};

char token[80];
char token_type, tok;

/* label lookup table */
struct label {
  char name[LAB_LEN]; /* label */
  char *p; /* points to label's location in source file */
} label_table[NUM_LAB];

/* support for FOR loops */
struct for_stack {
  int var; /* counter variable */
  int target;  /* target value */
  char *loc; /* place in source code to loop to */
} fstack[FOR_NEST]; /* stack for FOR/NEXT loop */

char *gstack[SUB_NEST];	/* stack for gosub */

int ftos;  /* index to top of FOR stack */
int gtos;  /* index to top of GOSUB stack */

void print(void), scan_labels(void), find_eol(void);
void exec_goto(void), exec_if(void), exec_for(void);
void next(void), fpush(struct for_stack i);
void input(void), gosub(void), greturn(void);
void gpush(char *s), label_init(void);
void assignment(void);
char *find_label(char *s), *gpop(void);
struct for_stack fpop(void);
int load_program(char *p, char *fname);
int get_next_label(char *s);

/* prototypes for functions in the parser file */
int eval_exp1(int *answer), get_token(void);
void serror(int error), putback(void);

/* These are the constants used to call serror() when
   a syntax error occurs.  Add more if you like.
   NOTE: SYNTAX is a generic error message used when
   nothing else seems appropriate.
*/
enum error_msg 
     {SYNTAX, UNBAL_PARENS, NO_EXP, EQUALS_EXP,
      NOT_VAR, LAB_TAB_FULL, DUP_LAB, UNDEF_LAB,
      THEN_EXP, TO_EXP, TOO_MNY_FOR, NEXT_WO_FOR,
      TOO_MNY_GOSUB, RET_WO_GOSUB, MISS_QUOTE};

main(int argc, char *argv[])
{
  if(argc!=2) {
    printf("usage: sbasic <filename>\n");
    exit(1);
  }

  /* allocate memory for the program */
  if(!(prog=(char *) malloc(PROG_SIZE))) {
    printf("allocation failure");
    exit(1);
  }
  p_buf = prog;
  
  /* load the program to execute */
  if(!load_program(prog, argv[1])) exit(1);

  if(setjmp(e_buf)) exit(1); /* initialize the long jump buffer */

  scan_labels(); /* find the labels in the program */
  ftos = 0; /* initialize the FOR stack index */
  gtos = 0; /* initialize the GOSUB stack index */
  do {
    token_type = get_token();
    /* check for assignment statement */
    if(token_type==VARIABLE) {
      putback(); /* return the var to the input stream */
      assignment(); /* must be assignment statement */
    }
    else /* is command */
      switch(tok) {
        case PRINT:
	  print();
  	  break;
        case GOTO:
	  exec_goto();
	  break;
	case IF:
	  exec_if();
	  break;
	case FOR:
	  exec_for();
	  break;
	case NEXT:
	  next();
	  break;
  	case INPUT:
	  input();
	  break;
        case GOSUB:
	  gosub();
	  break;
	case RETURN:
	  greturn();
	  break;
        case END:
	  return 0;
      }
  } while (tok != FINISHED);
  return 0;
}

/* Load a program. */
load_program(char *p, char *fname)
{
  FILE *fp;
  int i=0;

  if(!(fp=fopen(fname, "rb"))) return 0;
  
  i = 0;
  do {
    *p = getc(fp);
    p++; i++;
  } while(!feof(fp) && i<PROG_SIZE);
  *(p-2) = '\0'; /* null terminate the program */
  fclose(fp);
  return 1;
}

/* Find all labels. */
void scan_labels(void)
{
  int addr;
  char *temp;

  label_init();  /* zero all labels */
  temp = prog;   /* save pointer to top of program */

  /* if the first token in the file is a label */
  get_token();
  if(token_type==NUMBER) {
    strcpy(label_table[0].name, token);
    label_table[0].p = prog;
  }

  find_eol();
  do {     
    get_token();
    if(token_type==NUMBER) {
      addr = get_next_label(token);
      if(addr==-1 || addr==-2) {
        (addr==-1) ? serror(LAB_TAB_FULL):serror(DUP_LAB);
      }
      strcpy(label_table[addr].name, token);
     
      /* save current location in program */
      label_table[addr].p = prog;
    }
    /* if not on a blank line, find next line */
    if(tok!=EOL) find_eol();
  } while(tok!=FINISHED);
  prog = temp;  /* restore original location */
}

/* Find the start of the next line. */
void find_eol(void)
{
  while(*prog!='\n'  && *prog!='\0') ++prog;
  if(*prog) prog++;
}

/* Return index of next free position in label array. 
   -1 is returned if the array is full.
   -2 is returned when duplicate label is found.
*/
get_next_label(char *s)
{
  register int t;

  for(t=0;t<NUM_LAB;++t) {
    if(label_table[t].name[0]==0) return t;
    if(!strcmp(label_table[t].name,s)) return -2; /* dup */
  }

  return -1;
}

/* Find location of given label.  A null is returned if
   label is not found; otherwise a pointer to the position
   of the label is returned.
*/
char *find_label(char *s)
{
  register int t;

  for(t=0; t<NUM_LAB; ++t) 
    if(!strcmp(label_table[t].name,s)) return label_table[t].p;
  return '\0'; /* error condition */
}

/* Initialize the array that holds the labels. 
   By convention, a null label name indicates that
   array position is unused.
*/
void label_init(void)
{
  register int t;

  for(t=0; t<NUM_LAB; ++t) label_table[t].name[0]='\0';
}

/* Assign a variable a value. */
void assignment(void)
{
  int var, value;

  /* get the variable name */
  get_token();
  if(!isalpha(*token)) {
    serror(NOT_VAR);
    return;
  }

  /* convert to index into variable table */
  var = toupper(*token)-'A';
 
  /* get the equals sign */
  get_token();
  if(*token != '=') {
    serror(EQUALS_EXP);
    return;
  }

  /* get the value to assign */
  eval_exp1(&value);

  /* assign the value */
  variables[var] = value;
}

/* Execute a simple version of the BASIC PRINT statement. */
void print(void)
{
  int answer;
  int len=0, spaces;
  char last_delim;

  do {
    get_token(); /* get next list item */
    if(tok==EOL || tok==FINISHED) break;
    if(token_type==QUOTE) { /* is string */
      printf(token);
      len += strlen(token);
      get_token();
    }
    else { /* is expression */
      putback();
      eval_exp1(&answer);
      get_token();
      len += printf("%d", answer);
    }
    last_delim = *token; 

    /* if comma, move to next tab stop  */
    if(*token==',') {
      /* compute number of spaces to move to next tab */
      spaces = 8 - (len % 8); 
      len += spaces; /* add in the tabbing position */
      while(spaces) { 
	printf(" ");
        spaces--;
      }
    }
    else if(*token==';') printf(" ");
    else if(tok!=EOL && tok!=FINISHED) serror(SYNTAX); 
  } while (*token==';' || *token==',');

  if(tok==EOL || tok==FINISHED) {
    if(last_delim != ';' && last_delim!=',') printf("\n");
  }
  else serror(SYNTAX); /* error is not , or ; */

}

/* Execute a GOTO statement. */
void exec_goto(void)
{
  char *loc;

  get_token(); /* get label to go to */
  /* find the location of the label */
  loc = find_label(token);
  if(loc=='\0')
    serror(UNDEF_LAB); /* label not defined */

  else prog = loc;  /* start program running at that loc */
}

/* Execute an IF statement. */
void exec_if(void)
{
  int x, y, cond;
  char op;

  eval_exp1(&x); /* get left expression */

  get_token(); /* get the operator */
  if(!strchr("=<>", *token)) {
    serror(SYNTAX); /* not a legal operator */
    return;
  }
  op = *token;

  eval_exp1(&y); /* get right expression */

  /* determine the outcome */
  cond = 0;
  switch(op) {
    case '<':
      if(x<y) cond = 1;
      break;
    case '>':
      if(x>y) cond = 1;
      break;
    case '=':
      if(x==y) cond = 1;
      break;
  }
  if(cond) { /* is true so process target of IF */
    get_token();
    if(tok!=THEN) {
      serror(THEN_EXP);
      return;
    }/* else, target statement will be executed */
  }
  else find_eol(); /* find start of next line */
}

/* Execute a FOR loop. */
void exec_for(void)
{
  struct for_stack i;
  int value;

  get_token(); /* read the control variable */
  if(!isalpha(*token)) {
    serror(NOT_VAR);
    return;
  }

  i.var = toupper(*token)-'A'; /* save its index */

  get_token(); /* read the equals sign */
  if(*token!='=') {
    serror(EQUALS_EXP);
    return;
  }

  eval_exp1(&value); /* get initial value */

  variables[i.var] = value;

  get_token();
  if(tok!=TO) serror(TO_EXP); /* read and discard the TO */

  eval_exp1(&i.target); /* get target value */

  /* if loop can execute at least once, push info on stack */
  if(value >= variables[i.var]) { 
    i.loc = prog;
    fpush(i);
  }
  else  /* otherwise, skip loop code altogether */
    while(tok!=NEXT) get_token();
}

/* Execute a NEXT statement. */
void next(void)
{
  struct for_stack i;

  i = fpop(); /* read the loop info */

  variables[i.var]++; /* increment control variable */
  if(variables[i.var]>i.target) return;  /* all done */
  fpush(i);  /* otherwise, restore the info */
  prog = i.loc;  /* loop */
}

/* Push the FOR stack. */
void fpush(struct for_stack i)
{
   if(ftos>FOR_NEST)
    serror(TOO_MNY_FOR);

  fstack[ftos] = i;
  ftos++;
}

/* Pop the FOR stack. */
struct for_stack fpop(void)
{
  ftos--;
  if(ftos<0) serror(NEXT_WO_FOR);
  return(fstack[ftos]);
}

/* Execute a simple form of the BASIC INPUT command. */
void input(void)
{
  char var;
  int i;

  get_token(); /* see if prompt string is present */
  if(token_type==QUOTE) {
    printf(token); /* if so, print it and check for comma */
    get_token();
    if(*token!=',') serror(SYNTAX);
    get_token();
  }
  else printf("? "); /* otherwise, prompt with ? */
  var = toupper(*token)-'A'; /* get the input var */

  scanf("%d", &i); /* read input */

  variables[var] = i; /* store it */
}

/* Execute a GOSUB command. */
void gosub(void)
{
  char *loc;

  get_token();
  /* find the label to call */
  loc = find_label(token);
  if(loc=='\0')
    serror(UNDEF_LAB); /* label not defined */
  else {
    gpush(prog); /* save place to return to */
    prog = loc;  /* start program running at that loc */
  }
}

/* Return from GOSUB. */
void greturn(void)
{
   prog = gpop();
}

/* Push GOSUB stack. */
void gpush(char *s)
{
  gtos++;

  if(gtos==SUB_NEST) {
    serror(TOO_MNY_GOSUB);
    return;
  }

  gstack[gtos]=s;
}

/* Pop GOSUB stack. */
char *gpop(void)
{
  if(gtos==0) {
    serror(RET_WO_GOSUB);
    return 0;
  }

  return(gstack[gtos--]);
}

